from stub_generator.parse_docs import Docs


def test_multiple_declaration_with_docstrings():
    text = """
getEmpire() -> empire :
    Returns B

getEmpire( (int)arg1) -> empire :
    Returns A
""".strip(
        "\n"
    )

    docs = Docs(text, indent=0, is_class=False)
    assert list(docs.get_argument_strings()) == ["", "number: int"]
    assert docs.get_doc_string() == ('"""\nReturns B\nReturns A\n"""')


def test_single_declaration_with_docstrings():
    text = """
empirePlayerID( (int)arg1) -> int :
    Returns ...
    """.strip(
        "\n"
    )
    docs = Docs(text, indent=0, is_class=False)
    assert list(docs.get_argument_strings()) == ["number: int"]
    assert docs.get_doc_string() == '"""\nReturns ...\n"""'


def test_docs_single_line():
    text = "inQueue( (researchQueue)arg1, (str)arg2) -> bool"
    docs = Docs(text, indent=0, is_class=False)
    assert list(docs.get_argument_strings()) == ["research_queue: researchQueue, string: str"]
    assert docs.get_doc_string() == ""



from logging import debug

from common.statistic_interface import StatKey
from freeorion_tools.statistics.statistic import Stats


class StatsWrapper:
    def __init__(self):
        self._stats_writer = Stats(writer=debug)

    def _write(self, key, val):
        self._stats_writer.write(key, val)

    def ship_count(self, ship_count: int):
        self._write(StatKey.SHIP_CONT, ship_count)

    def capital(self, planet_id, planet_name, species):
        self._write(
            StatKey.CapitalID,
            {"capital_planet_id": planet_id, "capital_planet_name": planet_name, "capital_species": species},
        )

    def empire(self, id_, name, turn):
        return self._write(StatKey.EmpireID, {"empire_id": id_, "name": name, "turn": turn})

    def empire_color(self, r, g, b, a):
        return self._write(
            StatKey.EmpireColors,
            {
                "R": r,
                "G": g,
                "B": b,
                "A": a,
            },
        )

    def output(self, turn, rp, pp):
        self._write(StatKey.Output, {"turn": turn, "RP": rp, "PP": pp})

    def adopt_policy(self, name):
        self._write(StatKey.PolicyAdoption, name)

    def deadopt_policy(self, name):
        self._write(StatKey.PolicyDeAdoption, name)


stats = StatsWrapper()



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_TRANSLING_THT",
    description="LRN_TRANSLING_THT_DESC",
    short_description="POLICY_AND_BUILDING_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=32 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_LEARNING_CATEGORY", "THEORY"],
    prerequisites=["LRN_PHYS_BRAIN", "LRN_ALGO_ELEGANCE"],
    unlock=[
        Item(type=UnlockPolicy, name="PLC_MARINE_RECRUITMENT"),
        Item(type=UnlockPolicy, name="PLC_NATIVE_APPROPRIATION"),
        Item(type=UnlockBuilding, name="BLD_TRANSLATOR"),
    ],
    graphic="icons/tech/translingustic_thought.png",
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import TERRAN_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS
from species.species_macros.weapons import GREAT_WEAPONS
from species.species_macros.xenophobic import XENOPHOBIC_OTHER, XENOPHOBIC_SELF

Species(
    name="SP_EAXAW",
    description="SP_EAXAW_DESC",
    gameplay_description="SP_EAXAW_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ORGANIC",
        "XENOPHOBIC",
        "GREAT_WEAPONS",
        "AVERAGE_SUPPLY",
        "PEDIA_ORGANIC_SPECIES_CLASS",
        "PEDIA_XENOPHOBIC_SPECIES_TITLE",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "SHIMMER_SILK_SPECIAL",
        "FRACTAL_GEODES_SPECIAL",
        "SUCCULENT_BARNACLES_SPECIAL",
        "CRYSTALS_SPECIAL",
        "ANCIENT_RUINS_DEPLETED_SPECIAL",
        "PANOPTICON_SPECIAL",
        "PHILOSOPHER_SPECIAL",
        "PLC_CONTINUOUS_SCANNING",
        "PLC_TERRAFORMING",
        "PLC_CAPITAL_MARKETS",
        "PLC_DESIGN_SIMPLICITY",
        "PLC_CONFEDERATION",
    ],
    dislikes=[
        "PLC_DIVERSITY",
        "BLD_INTERSPECIES_ACADEMY",
        "BLD_XENORESURRECTION_LAB",
        "BLD_COLLECTIVE_NET",
        "BLD_LIGHTHOUSE",
        "BLD_REGIONAL_ADMIN",
        "RESONANT_MOON_SPECIAL",
        "TEMPORAL_ANOMALY_SPECIAL",
        "MINERALS_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "PLC_DIVERSITY",
        "PLC_ENVIRONMENTALISM",
        "PLC_ALLIED_REPAIR",
        "PLC_MODERATION",
        "PLC_NATIVE_APPROPRIATION",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *AVERAGE_RESEARCH,
        *AVERAGE_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_EAXAW"),
        *AVERAGE_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        *GREAT_WEAPONS,
        *XENOPHOBIC_SELF,
        XENOPHOBIC_OTHER("EAXAW"),
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=TERRAN_STANDARD_EP,
    graphic="icons/species/eaxaw.png",
)



from focs._effects import (
    Abs,
    Capital,
    EffectsGroup,
    EmpireHasAdoptedPolicy,
    Focus,
    GalaxyMaxAIAggression,
    Happiness,
    HasSpecies,
    HasTag,
    Homeworld,
    IsBuilding,
    IsHuman,
    IsSource,
    LocalCandidate,
    NamedReal,
    OwnedBy,
    Planet,
    ResourceSupplyConnected,
    SetTargetInfluence,
    Ship,
    Source,
    StatisticCount,
    Target,
    Turn,
    Unowned,
    Value,
)
from macros.misc import (
    DESCRIPTION_EFFECTSGROUP_MACRO,
    SUPPLY_DISCONNECTED_INFLUENCE_MALUS,
)
from macros.priorities import (
    TARGET_EARLY_BEFORE_SCALING_PRIORITY,
    TARGET_LAST_BEFORE_OVERRIDE_PRIORITY,
    TARGET_SCALING_PRIORITY,
)

_BASE_INFLUENCE_COSTS = [
    EffectsGroup(  # colonies consume influence, proportional to square-root of how many populated planets and non-populated outposts the empire controls
        scope=IsSource,
        activation=Planet() & ~Unowned & ~Capital,
        stackinggroup="IMPERIAL_PALACE_MANY_PLANETS_INFLUENCE_PENALTY",
        accountinglabel="COLONY_ADMIN_COSTS_LABEL",
        priority=TARGET_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetInfluence(
            value=Value
            - NamedReal(name="COLONY_ADMIN_COSTS_PER_PLANET", value=0.4)
            * (
                (
                    StatisticCount(
                        float,
                        condition=Planet()
                        & OwnedBy(empire=Source.Owner)
                        & HasSpecies()
                        & ~HasSpecies(name=["SP_EXOBOT"]),
                    )
                    + StatisticCount(
                        float,
                        condition=Ship & OwnedBy(empire=Source.Owner) & (LocalCandidate.OrderedColonizePlanetID != -1),
                    )
                    + StatisticCount(
                        float,
                        condition=IsBuilding()
                        & OwnedBy(empire=Source.Owner)
                        & IsBuilding(
                            name=[
                                "BLD_COL_SUPER_TEST",
                                "BLD_COL_ABADDONI",
                                "BLD_COL_BANFORO",
                                "BLD_COL_CHATO",
                                "BLD_COL_CRAY",
                                "BLD_COL_DERTHREAN",
                                "BLD_COL_EAXAW",
                                "BLD_COL_EGASSEM",
                                "BLD_COL_ETTY",
                                "BLD_COL_FULVER",
                                "BLD_COL_FURTHEST",
                                "BLD_COL_GEORGE",
                                "BLD_COL_GYSACHE",
                                "BLD_COL_HAPPY",
                                "BLD_COL_HHHOH",
                                "BLD_COL_HUMAN",
                                "BLD_COL_KILANDOW",
                                "BLD_COL_KOBUNTURA",
                                "BLD_COL_LAENFA",
                                "BLD_COL_MISIORLA",
                                "BLD_COL_MUURSH",
                                "BLD_COL_PHINNERT",
                                "BLD_COL_SCYLIOR",
                                "BLD_COL_SETINON",
                                "BLD_COL_SILEXIAN",
                                "BLD_COL_SLY",
                                "BLD_COL_SSLITH",
                                "BLD_COL_TAEGHIRUS",
                                "BLD_COL_TRITH",
                                "BLD_COL_REPLICON",
                                "BLD_COL_UGMORS",
                            ]
                        ),
                    )
                    + (
                        NamedReal(name="OUTPOST_RELATIVE_ADMIN_COUNT", value=0.25)
                        * (
                            StatisticCount(
                                float,
                                condition=Planet()
                                & OwnedBy(empire=Source.Owner)
                                & (~HasSpecies() | HasSpecies(name=["SP_EXOBOT"])),
                            )
                        )
                    )
                )
                ** 0.5
            )
        ),
    ),
    EffectsGroup(  # species homeworlds consume more influence
        scope=IsSource,
        activation=Planet() & ~Unowned & ~Capital & Homeworld(name=[LocalCandidate.Species]),
        accountinglabel="SPECIES_HOMEWORLD_INDEPENDENCE_DESIRE_LABEL",
        priority=TARGET_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetInfluence(value=Value - 1),
    ),
    EffectsGroup(  # colonies consume more influence if not supply connected to empire's capital
        scope=IsSource,
        activation=Planet()
        & ~Unowned
        & ~EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_CONFEDERATION")
        & ~ResourceSupplyConnected(
            empire=LocalCandidate.Owner, condition=Planet() & OwnedBy(empire=Source.Owner) & Capital
        )
        & Turn(low=1),
        accountinglabel="CAPITAL_DISCONNECTION_LABEL",
        priority=TARGET_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetInfluence(
            value=Value
            - NamedReal(name="SUPPLY_DISCONNECTED_INFLUENCE_MALUS", value=SUPPLY_DISCONNECTED_INFLUENCE_MALUS)
        ),
    ),
    EffectsGroup(  # gives human bonuses when AI Aggression set to Beginner
        scope=IsSource,
        activation=Planet() & ~Unowned & IsHuman & (GalaxyMaxAIAggression == 0),  # human player, not human species
        accountinglabel="DIFFICULTY",
        priority=TARGET_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetInfluence(value=Value + 1),
    ),
]

NO_INFLUENCE = [DESCRIPTION_EFFECTSGROUP_MACRO("NO_INFLUENCE_DESC"), *_BASE_INFLUENCE_COSTS]

_ARTISANS_INFLUENCE_STABILITY = [
    EffectsGroup(  # artistic species generate influence when artisans workshops policy adopted
        scope=IsSource,
        activation=HasTag(name="ARTISTIC")
        & Happiness(low=NamedReal(name="ARTISANS_MIN_STABILITY_FOCUS", value=1))
        & EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_ARTISAN_WORKSHOPS")
        & Focus(type=["FOCUS_INFLUENCE"]),
        accountinglabel="PLC_ARTISAN_WORKSHOPS",
        effects=SetTargetInfluence(value=Value + NamedReal(name="ARTISANS_INFLUENCE_FLAT_FOCUS", value=2.0)),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=HasTag(name="ARTISTIC")
        & Happiness(low=NamedReal(name="ARTISANS_MIN_STABILITY_NO_FOCUS", value=10))
        & EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_ARTISAN_WORKSHOPS")
        & ~Focus(type=["FOCUS_INFLUENCE"]),
        accountinglabel="PLC_ARTISAN_WORKSHOPS",
        effects=SetTargetInfluence(value=Value + NamedReal(name="ARTISANS_INFLUENCE_FLAT_NO_FOCUS", value=0.5)),
    ),
]


_BASIC_INFLUENCE = [
    EffectsGroup(  # influence focus generates influence from planets, proportional to sqare-root of population
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_INFLUENCE"]),
        accountinglabel="FOCUS_INFLUENCE_LABEL",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetTargetInfluence(
            value=Value
            + (Target.Population**0.5) * NamedReal(name="FOCUS_INFLUENCE_INFLUENCE_PER_SQRT_POP", value=1.0)
        ),
    ),
    *_BASE_INFLUENCE_COSTS,
    *_ARTISANS_INFLUENCE_STABILITY,
]


VERY_BAD_INFLUENCE = [
    *_BASIC_INFLUENCE,
    EffectsGroup(
        description="VERY_BAD_INFLUENCE_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_INFLUENCE"]),
        accountinglabel="VERY_BAD_INFLUENCE_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetInfluence(value=Value + (-0.33) * Abs(float, Value)),
    ),
]


BAD_INFLUENCE = [
    *_BASIC_INFLUENCE,
    EffectsGroup(
        description="BAD_INFLUENCE_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_INFLUENCE"]),
        accountinglabel="BAD_INFLUENCE_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetInfluence(value=Value + (-0.2) * Abs(float, Value)),
    ),
]


AVERAGE_INFLUENCE = _BASIC_INFLUENCE


GOOD_INFLUENCE = [
    *_BASIC_INFLUENCE,
    EffectsGroup(
        description="GOOD_INFLUENCE_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_INFLUENCE"]),
        accountinglabel="GOOD_INFLUENCE_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetInfluence(value=Value + (0.25) * Abs(float, Value)),
    ),
]

GREAT_INFLUENCE = [
    *_BASIC_INFLUENCE,
    EffectsGroup(
        description="GREAT_INFLUENCE_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_INFLUENCE"]),
        accountinglabel="GREAT_INFLUENCE_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetInfluence(value=Value + (0.5) * Abs(float, Value)),
    ),
]


ULTIMATE_INFLUENCE = [
    *_BASIC_INFLUENCE,
    EffectsGroup(
        description="ULTIMATE_INFLUENCE_DESC",
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_INFLUENCE"]),
        accountinglabel="ULTIMATE_INFLUENCE_LABEL",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetTargetInfluence(value=Value + (1.0) * Abs(float, Value)),
    ),
]



from focs._effects import (
    BuildBuilding,
    Capital,
    Conditional,
    ContainedBy,
    Contains,
    ContentFocus,
    CurrentTurn,
    EffectsGroup,
    EmpireMeterValue,
    EnemyOf,
    Enqueued,
    Fleet,
    GasGiantType,
    GenerateSitRepMessage,
    IsBuilding,
    IsSource,
    LocalCandidate,
    Location,
    Number,
    OneOf,
    OwnedBy,
    OwnerHasTech,
    Planet,
    RootCandidate,
    Ship,
    Source,
    Stationary,
    System,
    Target,
    Turn,
    TurnTechResearched,
    Unowned,
    UserString,
    VisibleToEmpire,
)
from macros.priorities import END_CLEANUP_PRIORITY

ETA_NEXT_TURN = CurrentTurn + LocalCandidate.ETA - 1

# The UserSitRepLabel tags and data are required if the message template (and label) is simply provided here rather than being drawn from the stringtable.
# The data value for the UserSitRepLabel serves as the label to be used in the SitReps filters, and Custom_1 through Custom_4
# will be at the top of the SitReps display.  As seen below, other labels are fine as well, but they cannot have any spaces if you want to use them in the SITREP_PRIORITY_ORDER
# More information on the scripting language used for these is available on our wiki;
# a suitable place to start is http:# www.freeorion.org/index.php/Free_Orion_Content_Script_(FOCS)


# Important, the following line "effectsgroups = [" is functional and important, do not change it
effectsgroups = [
    # normally, for a custom sitrep the player will write their own message here, and label, and not look them up in the Stringtable.  For this first custom sitrep,
    # however, since it is the primary way we introduce the custom sitrep capability to players, we use stringtable references so that the message and label may be more
    # readily translated into multiple languages as part of our standard distribution.
    EffectsGroup(
        scope=IsSource,
        activation=Turn(low=0, high=0),
        effects=GenerateSitRepMessage(
            message="CUSTOM_SITREP_INTRODUCTION",
            label="SITREP_WELCOME_LABEL",
            icon="icons/tech/categories/spy.png",
            parameters={"tech": "SPY_CUSTOM_ADVISORIES", "system": Target.SystemID},
            empire=Source.Owner,
        ),
    ),
    # for reference, the following is fully custom format (in english) of the intrduction sitrep above, except that it has a different label.
    # Note that the message and label are directly provided here rather than being stringtable keys,
    # and the NoStringtableLookup keyword is used afer the label is given.
    # If you are using a stringtable other than English, your translator may have translated the four preset labels for you,
    # so instead of the "Custom_1" below it might be something else for you.  The actual labels can be seen in the Pedia entry for
    # the SPY_CUSTOM_ADVISORIES tech (the tech is named "Empire Intelligence Agency" in the English version)
    #
    # EffectsGroup
    # scope = Source
    # activation = Turn low = 0 high = 0
    # effects =
    # GenerateSitRepMessage
    # message = "This is a sample Custom Sitrep, which is enabled by %tech% and all the busy empire servants reporting information back to %system%"
    # label = "Custom_1"
    # NoStringtableLookup
    # icon = "icons/tech/categories/spy.png"
    # parameters = [
    # tag = "tech" data = "SPY_CUSTOM_ADVISORIES"
    # tag = "system" data = Target.SystemID
    # ]
    # empire = Source.Owner
    #
    # EffectsGroup
    # scope = Source
    # activation = Turn low = 0 high = 0
    # effects =
    # GenerateSitRepMessage
    # message = "A number: %rawtext:number%  A boolean value: %rawtext:boolean%"
    # label = "Custom_1"
    # NoStringtableLookup
    # icon = "icons/tech/categories/spy.png"
    # parameters = [
    # tag = "number" data = GameRule name = "RULE_NUM_COMBAT_ROUNDS"
    # tag = "boolean" data = GameRule name = "RULE_CHEAP_AND_FAST_TECH_RESEARCH"
    # ]
    # empire = Source.Owner
    # Reminder if a Gas Giant you own in a system along with another planet you also own, populated by a species with Industry Focus, and you have the Orbital Generation tech
    # but haven't yet built or enqueued a GG Generator in that system yet
    EffectsGroup(
        scope=Planet()
        & Planet(type=[GasGiantType])
        & OwnedBy(empire=Source.Owner)
        & ~Contains(IsBuilding(name=["BLD_GAS_GIANT_GEN"]))
        & ContainedBy(
            Contains(
                Planet()
                & Location(type=ContentFocus, name=LocalCandidate.Species, name2="FOCUS_INDUSTRY")
                & OwnedBy(empire=Source.Owner)
            )
            & ~Contains(
                Planet()
                & OwnedBy(empire=Source.Owner)
                & (
                    Contains(IsBuilding(name=["BLD_GAS_GIANT_GEN"]))
                    | Enqueued(type=BuildBuilding, name="BLD_GAS_GIANT_GEN")
                )
            )
        ),
        activation=OwnerHasTech(name="PRO_ORBITAL_GEN"),
        effects=GenerateSitRepMessage(
            message="SITREP_GAS_GIANT_GENERATION_REMINDER",
            label="CUSTOM_2",
            icon="icons/building/gas-giant-generator.png",
            parameters={"buildingtype": "BLD_GAS_GIANT_GEN", "planet": Target.ID},
            empire=Source.Owner,
        ),
    ),
    # Tell all the others that you have better detection tech (the information is public anyway)
    EffectsGroup(
        scope=Capital & ~OwnedBy(empire=Source.Owner),
        activation=(CurrentTurn == TurnTechResearched(empire=Source.Owner, name="SPY_DETECT_2"))
        | (CurrentTurn == TurnTechResearched(empire=Source.Owner, name="SPY_DETECT_3"))
        | (CurrentTurn == TurnTechResearched(empire=Source.Owner, name="SPY_DETECT_4"))
        | (CurrentTurn == TurnTechResearched(empire=Source.Owner, name="SPY_DETECT_5")),
        priority=END_CLEANUP_PRIORITY,
        effects=GenerateSitRepMessage(
            message="SITREP_EMPIRE_TECH_RESEARCHED_DETECTION",
            label="SITREP_EMPIRE_TECH_RESEARCHED_DETECTION_LABEL",
            icon="icons/meter/detection.png",
            parameters={
                "empire": Source.Owner,
                "dstrength": EmpireMeterValue(empire=Source.Owner, meter="METER_DETECTION_STRENGTH"),
            },
            empire=Target.Owner,
        ),
    ),
    EffectsGroup(
        scope=System &
        # The following assets need protection
        Contains((Planet() | Ship) & OwnedBy(empire=Source.Owner)) &
        # Don't warn about system where there are already enemies this turn
        ~Contains(
            Ship & (OwnedBy(affiliation=EnemyOf, empire=Source.Owner) | Unowned) & VisibleToEmpire(empire=Source.Owner)
        ),
        activation=IsSource,
        priority=END_CLEANUP_PRIORITY,
        effects=Conditional(
            condition=Number(
                low=1,
                condition=Fleet
                & ~Stationary
                & (OwnedBy(affiliation=EnemyOf, empire=Source.Owner) | Unowned)
                & VisibleToEmpire(empire=Source.Owner)
                & Turn(low=ETA_NEXT_TURN, high=ETA_NEXT_TURN)
                & (RootCandidate.ID == LocalCandidate.NextSystemID),
            ),
            effects=[
                GenerateSitRepMessage(
                    message="SITREP_SYSTEM_GOT_INCOMING_WARNING",
                    label="SITREP_SYSTEM_GOT_INCOMING_WARNING_LABEL",
                    icon="icons/meter/ammo.png",
                    parameters={"system": Target.ID},
                    empire=Source.Owner,
                )
            ],
        ),
    ),
    # *********************************************************************************************************************************************************
    # The other example sitreps below are left in English, but can be translated just by changing the strings provided below for the messages and labels below.
    # It is important to note that, unlike the above entry, these entries all use the  NoStringtableLookup keyword after the label.
    # *********************************************************************************************************************************************************
    # Reminder if a Gas Giant you own in a system along with another planet you also own, populated by a species with Industry Focus, and you have the Orbital Generation tech
    # but haven't yet built or enqueued a GG Generator in that system yet
    # EffectsGroup
    # scope = And [
    # Planet
    # Planet Type = GasGiant
    # OwnedBy empire = Source.Owner
    # Not Contains Building name = "BLD_GAS_GIANT_GEN"
    # ContainedBy And [
    # Contains And [
    # Planet
    # Location type = Focus name = LocalCandidate.Species name = "FOCUS_INDUSTRY"
    # OwnedBy empire = Source.Owner
    # ]
    # Not Contains And [
    # Planet
    # OwnedBy empire = Source.Owner
    # Or [
    # Contains Building name = "BLD_GAS_GIANT_GEN"
    # Enqueued type = Building name = "BLD_GAS_GIANT_GEN"
    # ]
    # ]
    # ]
    # ]
    # activation = OwnerHasTech name = "PRO_ORBITAL_GEN"
    # effects =
    # GenerateSitRepMessage
    # message = "%planet% is a prime location for a %buildingtype%, but none is being built there"
    # label = "Custom_2"
    # NoStringtableLookup
    # icon = "icons/building/gas-giant-generator.png"
    # parameters = [
    # tag = "buildingtype" data = "BLD_GAS_GIANT_GEN"
    # tag = "planet" data = Target.ID
    # ]
    # empire = Source.Owner
    # Reminder if you own a planet with a monster nest but haven't yet researched SHP_DOMESTIC_MONSTER
    # EffectsGroup
    # scope = And [
    # Planet
    # OwnedBy empire = Source.Owner
    # Or [
    # HasSpecial name = "KRAKEN_NEST_SPECIAL"
    # HasSpecial name = "SNOWFLAKE_NEST_SPECIAL"
    # HasSpecial name = "JUGGERNAUT_NEST_SPECIAL"
    # ]
    # ]
    # activation = Not OwnerHasTech name = "SHP_DOMESTIC_MONSTER"
    # effects =
    # GenerateSitRepMessage
    # message = "You own %planet%, with a monster nest, but haven't yet researched %tech%"
    # label = "Custom_2"
    # NoStringtableLookup
    # icon = "icons/monsters/kraken-1.png"
    # parameters = [
    # tag = "planet" data = Target.ID
    # tag = "tech" data = "SHP_DOMESTIC_MONSTER"
    # ]
    # empire = Source.Owner
    # A Call to Arms
    # A prime candidate for being snoozed, but probably good to be periodically reminded about
    # EffectsGroup
    # scope = And [
    # Planet
    # Species
    # Unowned
    # MaxDefense high = 0
    # ContainedBy And [
    # System
    # Not Contains Monster
    # ]
    # VisibleToEmpire empire = Source.Owner
    # ]
    # activation = Source
    # effects =
    # GenerateSitRepMessage
    # message = "There are defenseless natives at %planet%!  Go invade them with [[metertype METER_TROOPS]]!"
    # label = "Custom_4"
    # NoStringtableLookup
    # icon = "icons/meter/troops.png"
    # parameters = [
    # tag = "planet" data = Target.ID
    # ]
    # empire = Source.Owner
    # #####     RANDOM BEGINNER HINTS     ######
    EffectsGroup(
        scope=IsSource,
        activation=Turn(low=0, high=100),
        stackinggroup="HINTSSYSTEM",
        effects=GenerateSitRepMessage(
            message="RANDOM_BEGINNER_HINT",
            label="BEGINNER_HINTS",
            icon="icons/sitrep/beginner_hint.png",
            parameters={
                "rawtext": OneOf(
                    str,
                    UserString("BEGINNER_HINT_01"),
                    UserString("BEGINNER_HINT_02"),
                    UserString("BEGINNER_HINT_03"),
                    UserString("BEGINNER_HINT_04"),
                    UserString("BEGINNER_HINT_05"),
                    UserString("BEGINNER_HINT_06"),
                    UserString("BEGINNER_HINT_07"),
                    UserString("BEGINNER_HINT_08"),
                    UserString("BEGINNER_HINT_09"),
                    UserString("BEGINNER_HINT_10"),
                    UserString("BEGINNER_HINT_11"),
                    UserString("BEGINNER_HINT_12"),
                    UserString("BEGINNER_HINT_13"),
                    UserString("BEGINNER_HINT_14"),
                    UserString("BEGINNER_HINT_15"),
                    UserString("BEGINNER_HINT_16"),
                    UserString("BEGINNER_HINT_17"),
                    UserString("BEGINNER_HINT_18"),
                    UserString("BEGINNER_HINT_19"),
                    UserString("BEGINNER_HINT_20"),
                    UserString("BEGINNER_HINT_21"),
                    UserString("BEGINNER_HINT_22"),
                )
            },
            empire=Source.Owner,
        ),
    )
    # Note, the "]" on the following line is functional and important; it must be present and all EffectsGroups must be above it.
]



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_MULTICELL_CAST",
    description="SHP_MULTICELL_CAST_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=20 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_ORGANIC_HULL_TECHS"],
    prerequisites=["SHP_ORG_HULL", "GRO_GENETIC_ENG"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_STATIC_MULTICELLULAR"),
        Item(type=UnlockBuilding, name="BLD_SHIPYARD_ORG_XENO_FAC"),
    ],
    graphic="icons/ship_hulls/static_multicellular_small.png",
)



from focs._effects import (
    Abs,
    EffectsGroup,
    EmpireHasAdoptedPolicy,
    HasEmpireStockpile,
    HasSpecies,
    Influence,
    IsSource,
    IsTarget,
    MaxOf,
    MinOf,
    NumPoliciesAdopted,
    OwnedBy,
    Planet,
    Population,
    SetConstruction,
    SetEmpireStockpile,
    SetHappiness,
    SetMaxTroops,
    SetTargetConstruction,
    Source,
    StatisticIf,
    Target,
    Value,
)
from focs._tech import *
from macros.priorities import (
    AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    METER_OVERRIDE_PRIORITY,
    TARGET_AFTER_2ND_SCALING_PRIORITY,
)

Tech(
    name="CON_OUTPOST",
    description="CON_OUTPOST_DESC",
    short_description="CON_OUTPOST",
    category="CONSTRUCTION_CATEGORY",
    researchcost=1,
    researchturns=1,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    unlock=Item(type=UnlockBuilding, name="BLD_ABANDON_OUTPOST"),
    # Effects for outposts
    effectsgroups=[
        # Outposts only have 50% of troops
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & Population(high=0)
            & ~EmpireHasAdoptedPolicy(empire=Source.Owner, name="PLC_MARINE_RECRUITMENT"),
            stackinggroup="OUTPOST_TROOPS_STACK",
            priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
            effects=SetMaxTroops(value=Value * 0.5),
            accountinglabel="OUTPOST_TROOP_LABEL",
        ),
        # Ensure construction minimum value of one, as this is necessary for being attacked
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            # has to happen after e.g. FORCE_ENERGY_STRC effects which also happens at AFTER_ALL_TARGET_MAX_METERS_PRIORITY
            priority=METER_OVERRIDE_PRIORITY,
            effects=[
                SetTargetConstruction(value=MaxOf(float, Value, 1)),
                SetConstruction(value=MaxOf(float, Value, 1)),
            ],
        ),
        # Influence growth / reduction towards target, since outposts have no species to get this effect from
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & (~HasSpecies()),
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetHappiness(
                value=Value
                + MinOf(float, Abs(float, Value(Target.TargetHappiness) - Value), 1)
                * (1 - 2 * (StatisticIf(float, condition=IsTarget & (Value > Value(Target.TargetHappiness)))))
            ),
        ),
        # Reset influence to 0 if no policies adopted. Not really relevant to Outposts, but I need somewhere to put this...
        EffectsGroup(
            scope=IsSource,
            activation=HasEmpireStockpile(empire=Source.Owner, resource=Influence, high=0)
            & (NumPoliciesAdopted(empire=Source.Owner) == 0),
            priority=METER_OVERRIDE_PRIORITY,
            effects=SetEmpireStockpile(resource=Influence, value=0.0),
        ),
    ],
)



from focs._effects import (
    EffectsGroup,
    Focus,
    OwnedBy,
    Planet,
    Population,
    SetMaxStockpile,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import STOCKPILE_PER_POP, TECH_COST_MULTIPLIER

Tech(
    name="PRO_VOID_PREDICTION",
    description="PRO_VOID_PREDICTION_DESC",
    short_description="IMPERIAL_STOCKPILE_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=700 * TECH_COST_MULTIPLIER,
    researchturns=7,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["LRN_MIND_VOID", "PRO_GENERIC_SUPPLIES"],
    unlock=Item(type=UnlockPolicy, name="PLC_STOCKPILE_LIQUIDATION"),
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & ~Population(high=0) & Focus(type=["FOCUS_STOCKPILE"]),
            effects=SetMaxStockpile(value=Value + 10 * Target.Population * STOCKPILE_PER_POP),
        )
    ],
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import DESERT_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import GOOD_INFLUENCE
from species.species_macros.native_fortification import ADVANCED_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import BAD_SUPPLY
from species.species_macros.troops import GREAT_DEFENSE_TROOPS
from species.species_macros.weapons import GREAT_WEAPONS

Species(
    name="SP_MUURSH",
    description="SP_MUURSH_DESC",
    gameplay_description="SP_MUURSH_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=["ORGANIC", "BAD_RESEARCH", "GOOD_INFLUENCE", "GREAT_WEAPONS", "BAD_SUPPLY", "PEDIA_ORGANIC_SPECIES_CLASS"],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INFLUENCE",
        "SUCCULENT_BARNACLES_SPECIAL",
        "MIMETIC_ALLOY_SPECIAL",
        "MINERALS_SPECIAL",
        "FRUIT_SPECIAL",
        "PLC_DIVERSITY",
        "PLC_INTERSTELLAR_INFRA",
        "PLC_NO_GROWTH",
        "PLC_PLANETARY_INFRA",
        "PLC_PROPAGANDA",
        "PLC_SYSTEM_INFRA",
        "PLC_TERROR_SUPPRESSION",
        "BLD_REGIONAL_ADMIN",
    ],
    dislikes=[
        "PLC_CONFEDERATION",
        "PLC_CONTINUOUS_SCANNING",
        "PLC_BLACK_MARKET",
        "PLC_INSURGENCY",
        "PLC_RACIAL_PURITY",
        "PLC_THE_HUNT",
        "BLD_SHIPYARD_ORBITAL_DRYDOCK",
        "BLD_SHIPYARD_ORG_ORB_INC",
        "BLD_SHIPYARD_AST",
        "BLD_SHIPYARD_CON_NANOROBO",
        "BLD_SHIPYARD_CON_ADV_ENGINE",
        "BLD_SHIPYARD_CON_GEOINT",
        "BLD_SHIPYARD_AST_REF",
        "BLD_SHIPYARD_ORG_XENO_FAC",
        "BLD_SHIPYARD_ENRG_COMP",
        "BLD_SHIPYARD_ENRG_SOLAR",
        "BLD_PLANET_CLOAK",
        "BLD_SPATIAL_DISTORT_GEN",
        "BLD_STARGATE",
        "BLD_MILITARY_COMMAND",
        "BLD_CONC_CAMP",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *BAD_RESEARCH,
        *GOOD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_MUURSH"),
        *BAD_SUPPLY,
        *GREAT_DEFENSE_TROOPS,
        *GREAT_WEAPONS,
        *ADVANCED_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=DESERT_STANDARD_EP,
    graphic="icons/species/muursh.png",
)



import freeorion as fo
import itertools
import random

import planets
import universe_statistics
import universe_tables

natives_for_planet_type = {}
planet_types_for_natives = {}


def generate_natives(native_freq, systems, empire_home_systems):  # noqa: C901
    """
    Adds non-empire-affiliated native populations to planets.
    And to asteroids and to gasgiants as well.
    """

    # first, calculate the chance for natives based on the native frequency that has been passed
    # get the corresponding value for the specified natives frequency from the universe tables
    native_chance = universe_tables.NATIVE_FREQUENCY[native_freq]
    # a value of 0 means no natives, in this case return immediately
    if native_chance <= 0:
        return

    # compile a list of systems where natives can be placed
    # select only systems sufficiently far away from player home systems
    # list of systems safe for natives
    EMPIRE_TO_NATIVE_MIN_DIST = 2
    empire_exclusions = set(
        itertools.chain.from_iterable(
            fo.systems_within_jumps_unordered(EMPIRE_TO_NATIVE_MIN_DIST, [e]) for e in empire_home_systems
        )
    )
    native_safe_planets = set(
        itertools.chain.from_iterable([fo.sys_get_planets(s) for s in systems if s not in empire_exclusions])
    )

    print("Number of planets far enough from players for natives to be allowed:", len(native_safe_planets))
    # if there are no "native safe" planets at all, we can stop here
    if not native_safe_planets:
        return

    # get all native species
    native_species = fo.get_native_species()
    print("Species that can be added as natives:")
    print("... " + "\n... ".join(native_species))

    # create a map with a list for each planet type containing the species
    # for which this planet type is a good environment
    # we will need this afterwards when picking natives for a planet
    natives_for_planet_type.clear()  # just to be safe
    natives_for_planet_type.update({planet_type: [] for planet_type in planets.planet_types})
    planet_types_for_natives.clear()
    planet_types_for_natives.update({species: set() for species in native_species})

    # iterate over all native species we got
    for species in native_species:
        # check the planet environment for all planet types for this species
        for planet_type in planets.planet_types:
            # if this planet type is a good environment for the species, add it to the list for this planet type
            if fo.species_get_planet_environment(species, planet_type) == fo.planetEnvironment.good:
                natives_for_planet_type[planet_type].append(species)
                planet_types_for_natives[species].add(planet_type)
        # if the species simply has no good environment to live in, try the same as above for adequate instead
        # this is needed for a couple of new native species with only adequate and no good environments
        if len(planet_types_for_natives[species]) < 1:
            for planet_type in planets.planet_types:
                if fo.species_get_planet_environment(species, planet_type) == fo.planetEnvironment.adequate:
                    natives_for_planet_type[planet_type].append(species)
                    planet_types_for_natives[species].add(planet_type)
        # if the species still has no fitting environment then try poor next
        if len(planet_types_for_natives[species]) < 1:
            for planet_type in planets.planet_types:
                if fo.species_get_planet_environment(species, planet_type) == fo.planetEnvironment.poor:
                    natives_for_planet_type[planet_type].append(species)
                    planet_types_for_natives[species].add(planet_type)

    # randomly add species to planets
    # iterate over the list of "native safe" planets we compiled earlier
    for candidate in native_safe_planets:
        # select a native species to put on this planet
        planet_type = fo.planet_get_type(candidate)
        # check if we have any native species that like this planet type
        if not natives_for_planet_type[planet_type]:
            # no, continue with next planet
            continue
        universe_statistics.potential_native_planet_summary[planet_type] += 1
        # temporarily dampen native frequencies on roids and GGs until better galaxy setup options are in place
        factor = 1
        if planet_type in [fo.planetType.asteroids, fo.planetType.gasGiant]:
            factor = 5
        # make a "roll" against the chance for natives to determine if we shall place natives on this planet
        if random.random() > native_chance / factor:
            # no, continue with next planet
            continue
        universe_statistics.settled_native_planet_summary[planet_type] += 1

        # randomly pick one of the native species available for this planet type
        natives = random.choice(natives_for_planet_type[planet_type])

        # put the selected natives on the planet
        fo.planet_set_species(candidate, natives)
        # set planet as homeworld for that species
        fo.species_add_homeworld(natives, candidate)
        # set planet focus
        # check if the preferred focus for the native species is among the foci available on this planet
        available_foci = fo.planet_available_foci(candidate)
        preferred_focus = fo.species_preferred_focus(natives)
        if preferred_focus in available_foci:
            # if yes, set the planet focus to the preferred focus
            fo.planet_set_focus(candidate, preferred_focus)
        elif available_foci:
            # if no, and there is at least one available focus, just take the first of the list
            # otherwise don't set any focus
            fo.planet_set_focus(candidate, available_foci[0])
            print("preferred focus", preferred_focus, "not available, focus on", available_foci[0], "instead")
        print("Added native", natives, "to planet", fo.get_name(candidate))

        # increase the statistics counter for this native species, so a species summary can be dumped to the log later
        universe_statistics.species_summary[natives] += 1



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from techs.ship_weapons.ship_weapons import WEAPON_BASE_EFFECTS

Tech(
    name="SHP_WEAPON_2_1",
    description="SHP_WEAPON_2_1_DESC",
    short_description="SHIP_WEAPON_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=64 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_SR_WEAPON_TECHS"],
    prerequisites=["SHP_ROOT_AGGRESSION"],
    unlock=Item(type=UnlockShipPart, name="SR_WEAPON_2_1"),
    effectsgroups=WEAPON_BASE_EFFECTS("SR_WEAPON_2_1"),
    graphic="icons/ship_parts/laser-1.png",
)



from focs._effects import EffectsGroup, GiveEmpireTech, IsSource, Target, Turn
from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import BROAD_EP, INFERNO_BROAD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import GOOD_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import GREAT_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS

Species(
    name="SP_ABADDONI",
    description="SP_ABADDONI_DESC",
    gameplay_description="SP_ABADDONI_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "LITHIC",
        "BAD_RESEARCH",
        "GREAT_INFLUENCE",
        "GOOD_HAPPINESS",
        "AVERAGE_SUPPLY",
        "PEDIA_LITHIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "SHIMMER_SILK_SPECIAL",
        "FRACTAL_GEODES_SPECIAL",
        "SUPERCONDUCTOR_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "MINERALS_SPECIAL",
        "CRYSTALS_SPECIAL",
        "PLC_DIVINE_AUTHORITY",
        "PLC_CONFORMANCE",
        "PLC_TERROR_SUPPRESSION",
        "PLC_INDOCTRINATION",
        "PLC_INSURGENCY",
        "PLC_BUREAUCRACY",
    ],
    dislikes=[
        "BLD_SCRYING_SPHERE",
        "BLD_MEGALITH",
        "BLD_PLANET_DRIVE",
        "BLD_GATEWAY_VOID",
        "BLD_GAS_GIANT_GEN",
        "FORTRESS_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "PHILOSOPHER_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "PLC_DIVERSITY",
        "PLC_FEUDALISM",
        "PLC_LIBERTY",
        "PLC_ARTISAN_WORKSHOPS",
        "PLC_CONFEDERATION",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *BAD_RESEARCH,
        *GREAT_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *GOOD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_ABADDONI"),
        *AVERAGE_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        EffectsGroup(
            description="ABADDONI_STARTING_UNLOCKS_DESC",
            scope=IsSource,
            activation=Turn(high=1),
            effects=[GiveEmpireTech(name="GRO_SUBTER_HAB", empire=Target.Owner)],
        ),
        # not for description
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        BROAD_EP,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=INFERNO_BROAD_EP,
    graphic="icons/species/abaddonnian.png",
)



from collections.abc import Mapping, Sequence
from logging import info

from common.fo_typing import PlanetId
from empire.survey_lock import survey_universe_lock
from freeorion_tools.caching import cache_for_current_turn


def set_pilot_rating_for_planet(pid: PlanetId, pilot_rating: float):
    """
    Set pilot rating for planet.

    Warning! Temporal coupling.
    All calls of this function should be done before using of this information.
    """
    _get_pilot_ratings()[pid] = pilot_rating


@survey_universe_lock
def get_rating_for_planet(pid: PlanetId) -> float:
    return get_pilot_ratings().get(pid, 0)


@survey_universe_lock
def get_pilot_ratings() -> Mapping[PlanetId, float]:
    return _get_pilot_ratings()


@cache_for_current_turn
def _get_pilot_ratings() -> dict[PlanetId, float]:
    """
    Return mutable state.
    """
    return {}


class _Summary:
    def __init__(self):
        self.best_pilot_rating = 1e-8
        self.medium_pilot_rating = 1e-8


@cache_for_current_turn
def _get_pilot_summary() -> _Summary:
    summary = _Summary()
    rating_list = sorted(get_pilot_ratings().values(), reverse=True)
    _summarize_pilot_ratings(summary, rating_list)
    return summary


@survey_universe_lock
def best_pilot_rating() -> float:
    return _get_pilot_summary().best_pilot_rating


@survey_universe_lock
def medium_pilot_rating() -> float:
    return _get_pilot_summary().medium_pilot_rating


def _summarize_pilot_ratings(summary: _Summary, ratings: Sequence[float]):
    if not ratings:
        return
    info("_summarize_pilot_ratings")
    summary.best_pilot_rating = ratings[0]
    if len(ratings) == 1:
        summary.medium_pilot_rating = ratings[0]
    else:
        summary.medium_pilot_rating = ratings[(1 + int(len(ratings) // 5))]



import pytest

import savegame_codec
from CombatRatingsAI import ShipCombatStats
from savegame_codec._definitions import TrustedClasses


class DummyTestClass:
    def __init__(self):
        self.some_int = 812
        self.some_negative_int = -712
        self.some_float = 1.2
        self.some_other_float = 1.248e3
        self.some_string = "TestString"
        self._some_private_var = "_Test"
        self.__some_more_private_var = "__Test"
        self.some_dict = {"ABC": 123.1, (1, 2.3, "ABC", "zz"): (1, (2, (3, 4)))}
        self.some_list = [self.some_int, self.some_float, self.some_other_float, self.some_string]
        self.some_set = set(self.some_list)
        self.some_tuple = tuple(self.some_list)

    def dummy_class_function(self):
        pass


@pytest.fixture(
    params=[
        0,
        0.0,
        "",
        False,
        1,
        1.2,
        1.2e4,
        "TestString",
        True,
        False,
        None,
        tuple(),
        set(),
        list(),
        dict(),
        (1, 2, 3),
        [1, 2, 3],
        {1, 2, 3},
        {"a": 1, True: False, (1, 2): {1, 2, 3}, (1, 2, (3, (4,))): [1, 2, (3, 4), {1, 2, 3}]},
    ],
    ids=str,
)
def simple_object(
    request,
):
    return request.param


def test_class_with_protected_attribute():
    foo = ShipCombatStats()
    trusted_classes = TrustedClasses()
    retval = savegame_codec.encode(foo, TrustedClasses())
    assert retval
    assert isinstance(retval, str)

    restored_obj = savegame_codec.decode(retval, trusted_classes)
    assert type(restored_obj) is type(foo)
    assert restored_obj == foo


def check_encoding(obj):
    trusted_classes = TrustedClasses()
    retval = savegame_codec.encode(obj, trusted_classes)
    assert retval
    assert isinstance(retval, str)

    restored_obj = savegame_codec.decode(retval, trusted_classes)
    assert type(restored_obj) is type(obj)
    assert restored_obj == obj


def test_encoding_simple_object(simple_object):
    check_encoding(simple_object)


def test_encoding_function():
    match = "Class builtins.function is not trusted"

    with pytest.raises(savegame_codec.CanNotSaveGameException, match=match):
        check_encoding(lambda: 0)
        pytest.fail("Could save function")


def test_encoding_type():
    match = "Class builtins.type is not trusted$"

    with pytest.raises(savegame_codec.CanNotSaveGameException, match=match):
        check_encoding(list)
        pytest.fail("Could save untrusted class")


def test_class_encoding():
    trusted_scope = TrustedClasses()
    obj = DummyTestClass()
    with pytest.raises(
        savegame_codec.CanNotSaveGameException, match="Class test_savegame_manager.DummyTestClass is not trusted"
    ):
        savegame_codec.encode(obj, trusted_scope)
        pytest.fail("Could save game even though test module should not be trusted")

    new_trusted_scope = TrustedClasses({__name__ + ".DummyTestClass": DummyTestClass})

    retval = savegame_codec.encode(obj, new_trusted_scope)

    assert retval
    assert isinstance(retval, str)

    loaded_obj = savegame_codec.decode(retval, new_trusted_scope)
    assert isinstance(loaded_obj, DummyTestClass)
    assert type(loaded_obj) is type(obj)

    original_dict = loaded_obj.__dict__
    loaded_dict = loaded_obj.__dict__

    assert loaded_dict == original_dict


def test_enums():
    import EnumsAI

    test_obj = {EnumsAI.MissionType.COLONISATION: EnumsAI.MissionType.INVASION}
    trusted_classes = TrustedClasses()
    restored_obj = savegame_codec.decode(savegame_codec.encode(test_obj, TrustedClasses()), trusted_classes)
    assert test_obj == restored_obj



from focs._effects import (
    AddSpecial,
    AsteroidsType,
    Contains,
    Destroy,
    EffectsGroup,
    HasSpecial,
    IsBuilding,
    IsSource,
    Object,
    OwnedBy,
    Planet,
    Source,
)
from macros.base_prod import BUILDING_COST_MULTIPLIER
from macros.enqueue import ENQUEUE_BUILD_ONE_PER_PLANET

try:
    from focs._buildings import *
except ModuleNotFoundError:
    pass

BuildingType(  # type: ignore[reportUnboundVariable]
    name="BLD_ART_MOON",
    description="BLD_ART_MOON_DESC",
    buildcost=250 * BUILDING_COST_MULTIPLIER,
    buildtime=5,
    location=(
        Planet()
        & ~Contains(IsBuilding(name=["BLD_ART_MOON"]))
        & OwnedBy(empire=Source.Owner)
        & ~Planet(type=[AsteroidsType])
        & ~HasSpecial(name="RESONANT_MOON_SPECIAL")
    ),
    enqueuelocation=ENQUEUE_BUILD_ONE_PER_PLANET,
    effectsgroups=[
        EffectsGroup(
            scope=Object(id=Source.PlanetID) & Planet(),
            effects=AddSpecial(name="RESONANT_MOON_SPECIAL"),
        ),
        EffectsGroup(scope=IsSource, effects=Destroy),
    ],
    icon="icons/specials_huge/resonant_moon.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_KRILL_SPAWN",
    description="SHP_KRILL_SPAWN_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=9999 * TECH_COST_MULTIPLIER,
    researchturns=9999,
    researchable=False,
    tags=["PEDIA_SHIP_WEAPONS_CATEGORY"],
    unlock=Item(type=UnlockShipPart, name="SP_KRILL_SPAWNER"),
)



import freeOrionAIInterface as fo
from math import sqrt

from aistate_interface import get_aistate
from buildings import BuildingType
from EnumsAI import FocusType, PriorityType
from freeorion_tools import get_named_real, get_species_influence
from turn_state import get_empire_planets_by_species

INFLUENCE_SPECIALS = [
    "FRACTAL_GEODES_SPECIAL",
    "MIMETIC_ALLOY_SPECIAL",
    "SHIMMER_SILK_SPECIAL",
    "SPARK_FOSSILS_SPECIAL",
    "SUCCULENT_BARNACLES_SPECIAL",
]
SPECIAL_FLAT = get_named_real("SPECIAL_INFLUENCE_FOCUS_BONUS")
ARTISAN_FLAT = get_named_real("ARTISANS_INFLUENCE_FLAT_FOCUS")
PLC_ARTISAN = "PLC_ARTISAN_WORKSHOPS"
TAG_ARTISTIC = "ARTISTIC"


def calculate_influence(planet: fo.planet, species: fo.species, max_pop: float, stability: float) -> float:
    """
    Calculate how much influence the planet could generate with influence focus.
    """
    if not species or FocusType.FOCUS_INFLUENCE not in species.foci:
        return 0.0
    production = sqrt(max_pop)
    production += sum(SPECIAL_FLAT for special in planet.specials if special in INFLUENCE_SPECIALS)
    if fo.getEmpire().policyAdopted(PLC_ARTISAN) and TAG_ARTISTIC in species.tags:
        offset = stability - get_named_real("ARTISANS_MIN_STABILITY_FOCUS")
        production += ARTISAN_FLAT / (1.0 if offset >= 0 else offset**2 + 2)  # grant a little, if close
    # So far all those flat bonuses are affected by the species multiplier
    production *= get_species_influence(species.name)

    translator = BuildingType.TRANSLATOR
    bonus = get_named_real("TRANSLATOR_PER_SPECIES_INFLUENCE") * len(get_empire_planets_by_species()) ** 0.5
    # We may consider conquering a planet that has one. Note that translator.build_at shows only our translators
    if any(translator.is_this_type(bld_id) for bld_id in planet.buildingIDs):
        production += bonus
    elif translator.available():  # we'd have to build it first, so use only half of the bonus
        production += 0.5 * production
    # TODO Species InterDesign Academy
    return production


def rate_influence(planet, species, max_pop):  # TODO: remove with the planet rating old code...
    """
    Give a rating how useful the colony would be for generating influence.
    This should be used as an alternative for industry/research.
    """
    production = calculate_influence(planet, species, max_pop, 10.0)
    # TBD check for nearly universal translator:
    # 1. We could rate a planet that already has one
    # 2. The ability to build one could remove the -1 below
    # Other factors still missing: policies Terraforming, Indoctrination
    # and Environmentalism (most likely never used by AI).
    influence_priority = get_aistate().get_priority(PriorityType.RESOURCE_INFLUENCE)
    # a production of 1 is hardly worth anything
    return (15 + influence_priority / 10) * (production - 1)



import freeOrionAIInterface as fo
from collections import namedtuple
from heapq import heappop, heappush
from logging import error, warning
from typing import Callable, Optional

import PlanetUtilsAI
from AIDependencies import INVALID_ID
from aistate_interface import get_aistate
from common.fo_typing import SystemId
from EnumsAI import MissionType
from freeorion_tools import chat_human, get_partial_visibility_turn
from freeorion_tools.caching import cache_for_current_turn
from turn_state import get_system_supply
from universe.system_network import get_neighbors, get_shortest_distance

_DEBUG_CHAT = False
_ACCEPTABLE_DETOUR_LENGTH = 2000
PathInformation = namedtuple("PathInformation", ["distance", "fuel", "path"])


# cache this so that boost python does not need to make a new copy every time this info is needed in a turn.
@cache_for_current_turn
def _get_unobstructed_systems():
    return fo.getEmpire().supplyUnobstructedSystems


def _info_string(path_info):
    return f"dist {path_info.distance:.1f}, path {PlanetUtilsAI.sys_name_ids(path_info.path)}"


# Note that this can cover departure from uncontested systems as well as from contested systems where our forces
# arrived first (and those forces may or may not be the forces for whom the pathfinding is being done)
def _more_careful_travel_starlane_func(c, d):
    return c in _get_unobstructed_systems()


# In comparison to  _more_careful_travel_starlane_func, this can also allow passage through contested
# systems where our forces already present there (which may or may not be the forces for whome the pathfinding
# is being done) arrived second but are currently preserving travel along the starlane of interest.
def _somewhat_careful_travel_starlane_func(c, d):
    return any((c in _get_unobstructed_systems(), fo.getEmpire().preservedLaneTravel(c, d)))


# For some activities like scouting, we may want to allow an extra bit of risk from routing through a system which
# has all its exits necessarily marked as restricted simply because it has never been partially visible
def _risky_travel_starlane_func(c, d):
    return any((_somewhat_careful_travel_starlane_func(c, d), get_partial_visibility_turn(c) <= 0))


def _may_travel_anywhere(*args, **kwargs):
    return True


_STARLANE_TRAVEL_FUNC_MAP = {
    MissionType.EXPLORATION: _risky_travel_starlane_func,
    MissionType.OUTPOST: _more_careful_travel_starlane_func,
    MissionType.COLONISATION: _more_careful_travel_starlane_func,
    MissionType.INVASION: _somewhat_careful_travel_starlane_func,
    MissionType.MILITARY: _may_travel_anywhere,
    MissionType.SECURE: _may_travel_anywhere,
}


def find_path_with_resupply(
    start: int,
    target: int,
    fleet_id: int,
    minimum_fuel_at_target: int = 0,
    mission_type_override: Optional[MissionType] = None,
) -> PathInformation:
    """
    :param start: start system id
    :param target:  target system id
    :param fleet_id: fleet to find the path for
    :param minimum_fuel_at_target: optional - if specified, only accept paths that leave the
                                   fleet with at least this much fuel left at the target system
    :param mission_type_override: optional - use the specified mission type, rather than the fleet's
                                  current mission type, for pathfinding routing choices
    :return: shortest possible path including resupply-detours in the form of system ids
             including both start and target system
    """

    universe = fo.getUniverse()
    fleet = universe.getFleet(fleet_id)
    if not fleet:
        return None
    empire = fo.getEmpire()
    supplied_systems = set(empire.fleetSupplyableSystemIDs)
    start_fuel = fleet.maxFuel if start in supplied_systems else fleet.fuel

    # We have 1 free jump from supplied system into unsupplied systems.
    # Thus, the target system must be at most maxFuel + 1 jumps away
    # in order to reach the system under standard conditions.
    # In some edge cases, we may have more supply here than what the
    # supply graph suggests. For example, we could have recently refueled
    # in a system that is now blockaded by an enemy or we have found
    # the refuel special.
    target_distance_from_supply = -min(get_system_supply(target), 0)
    if fleet.maxFuel + 1 < (target_distance_from_supply + minimum_fuel_at_target) and universe.jumpDistance(
        start, target
    ) > (start_fuel - minimum_fuel_at_target):
        # can't possibly reach this system with the required fuel
        return None

    mission_type = (
        mission_type_override if mission_type_override is not None else get_aistate().get_fleet_mission(fleet_id)
    )
    may_travel_starlane_func = _STARLANE_TRAVEL_FUNC_MAP.get(mission_type, _more_careful_travel_starlane_func)

    path_info = find_path_with_resupply_generic(
        start,
        target,
        start_fuel,
        fleet.maxFuel,
        lambda s: s in supplied_systems,
        minimum_fuel_at_target,
        may_travel_starlane_func=may_travel_starlane_func,
    )

    if not _DEBUG_CHAT:
        return path_info

    if may_travel_starlane_func != _may_travel_anywhere:
        risky_path = find_path_with_resupply_generic(
            start, target, start_fuel, fleet.maxFuel, lambda s: s in supplied_systems, minimum_fuel_at_target
        )
        if path_info and may_travel_starlane_func == _risky_travel_starlane_func:
            safest_path = find_path_with_resupply_generic(
                start,
                target,
                start_fuel,
                fleet.maxFuel,
                lambda s: s in supplied_systems,
                minimum_fuel_at_target,
                may_travel_starlane_func=_more_careful_travel_starlane_func,
            )
            if safest_path and path_info.distance < safest_path.distance:
                message = "(Scout?) Fleet %d chose somewhat risky path %s instead of safe path %s" % (
                    fleet_id,
                    _info_string(path_info),
                    _info_string(risky_path),
                )
                chat_human(message)

        if path_info and risky_path and risky_path.distance < path_info.distance:
            message = "Fleet %d chose safer path %s instead of risky path %s" % (
                fleet_id,
                _info_string(path_info),
                _info_string(risky_path),
            )
            chat_human(message)

    return path_info


# TODO Add a list of systems as parameter that are forbidden for search (e.g. blocked by monster)
# TODO Allow short idling in system to use stationary refuel mechanics in unsupplied systems if we
#      are very close to the target system but can't reach it due to fuel constraints
# TODO Consider additional optimizations:
#    - Check if universe.shortestPath complies with fuel mechanics before running
#      pathfinder (seemingly not worth it - we have exact heuristic in that case)
#    - Cut off branches that can't reach target due to supply distance
#    - For large graphs, check existence of a path on a simplified graph (use single node to represent supply clusters)
#    - For large graphs, check if there are any must-visit nodes (e.g. only possible resupplying system),
#      then try to find the shortest path between those and start/target.
def find_path_with_resupply_generic(  # noqa: C901
    start: SystemId,
    target: SystemId,
    start_fuel: float,
    max_fuel: float,
    system_suppliable_func: Callable[[int], bool],
    minimum_fuel_at_target=0,
    may_travel_system_func: Optional[Callable[[int], bool]] = None,
    may_travel_starlane_func: Optional[Callable[[int, int], bool]] = None,
) -> Optional[PathInformation]:
    """Find the shortest possible path between two systems that complies with FreeOrion fuel mechanics.

     If the fleet can travel the shortest possible path between start and target system, then return that path.
     Otherwise, find the shortest possible detour including refueling.

     The core algorithm is a modified A* with the universe.shortestPathDistance as heuristic.
     While searching for a path, keep track of the fleet's fuel. Compared to standard A*/dijkstra,
     nodes are locked only for a certain minimum level of fuel - if a longer path yields a higher fuel
     level at a given system, then that path is considered as possible detour for refueling and added to the queue.

    :param start_fuel: starting fuel of the fleet
    :param max_fuel: max fuel of the fleet
    :param system_suppliable_func: boolean function with one int param s, specifying if a system s provides fleet supply
    :param minimum_fuel_at_target: optional - if specified, only accept paths that leave the
                                   fleet with at least this much fuel left at the target system
    :param may_travel_system_func: optional - boolean function with one int param, s, specifying if
                                   a system s is OK to travel through
    :param may_travel_starlane_func: optional - boolean function with 2 int params c, d, specifying if
                                     a starlane from c to d is OK to travel through
    :return: shortest possible path including resupply-detours in the form of system ids
             including both start and target system
    """

    universe = fo.getUniverse()

    if start == INVALID_ID or target == INVALID_ID:
        warning("Requested path between invalid systems.")
        return None

    # make sure the minimum fuel at target is realistic
    if minimum_fuel_at_target < 0:
        error("Requested negative fuel at target.")
        return None

    # make sure the minimum fuel at target is realistic
    if max_fuel < minimum_fuel_at_target:
        return None

    # make sure the target is connected to the start system
    shortest_possible_path_distance = get_shortest_distance(start, target)
    if shortest_possible_path_distance == -1:
        warning("Requested path between disconnected systems, doing nothing.")
        return None

    if may_travel_system_func is None:
        may_travel_system_func = _may_travel_anywhere
    if may_travel_starlane_func is None:
        may_travel_starlane_func = _may_travel_anywhere

    # initialize data structures
    path_cache = {}
    queue = []

    # add starting system to queue
    heappush(queue, (shortest_possible_path_distance, PathInformation(distance=0, fuel=start_fuel, path=(start,))))

    while queue:
        # get next system with path information
        (_, path_info) = heappop(queue)
        current = path_info.path[-1]

        # did we reach the target?
        if current == target:
            # do we satisfy fuel constraints?
            if path_info.fuel < minimum_fuel_at_target:
                continue
            return path_info

        # add information about how we reached here to the cache
        path_cache.setdefault(current, []).append(path_info)

        # check if we have enough fuel to travel to neighbors
        if path_info.fuel < 1:
            continue

        # add neighboring systems to the queue if the resulting path
        # is either shorter or offers more fuel than the other paths
        # which we already found to those systems
        for neighbor in get_neighbors(current):
            # A system we have never had partial vis for will count as fully blockaded for us, but perhaps if
            # we are scouting we might want to be able to route a path through it anyway.
            if any(
                (
                    not may_travel_starlane_func(current, neighbor),
                    neighbor != target and not may_travel_system_func(neighbor),
                )
            ):
                continue
            new_dist = path_info.distance + universe.linearDistance(current, neighbor)
            new_fuel = (
                max_fuel
                if (system_suppliable_func(neighbor) or system_suppliable_func(current))
                else path_info.fuel - 1
            )

            # check if the node is already closed, i.e. a path was already found which both is shorter and offers
            # more fuel. Priority queueing should ensure that all previously found paths here are shorter but check
            # it anyway to be sure...
            if any((dist <= new_dist and fuel >= new_fuel) for dist, fuel, _ in path_cache.get(neighbor, [])):
                continue

            # calculate the new distance prediction, i.e. the A* heuristic.
            predicted_distance = new_dist + get_shortest_distance(neighbor, target)

            # Ignore paths that are much longer than the shortest possible path
            if predicted_distance > max(
                2 * shortest_possible_path_distance, shortest_possible_path_distance + _ACCEPTABLE_DETOUR_LENGTH
            ):
                continue

            # All checks passed, consider this path for further pathfinding
            heappush(queue, (predicted_distance, PathInformation(new_dist, new_fuel, path_info.path + (neighbor,))))

    # no path exists, not even if we refuel on the way
    return None



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import TUNDRA_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import GOOD_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import GOOD_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import GREAT_DEFENSE_TROOPS

Species(
    name="SP_HHHOH",
    description="SP_HHHOH_DESC",
    gameplay_description="SP_HHHOH_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=["ORGANIC", "GOOD_INFLUENCE", "GOOD_HAPPINESS", "AVERAGE_SUPPLY", "PEDIA_ORGANIC_SPECIES_CLASS"],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "ECCENTRIC_ORBIT_SPECIAL",
        "FOCUS_INDUSTRY",
        "KRAKEN_NEST_SPECIAL",
        "MINERALS_SPECIAL",
        "PLC_CONFORMANCE",
        "PLC_MODERATION",
        "RESONANT_MOON_SPECIAL",
        "SHIMMER_SILK_SPECIAL",
        "SUCCULENT_BARNACLES_SPECIAL",
    ],
    dislikes=[
        "BLD_COLLECTIVE_NET",
        "BLD_SCANNING_FACILITY",
        "ELERIUM_SPECIAL",
        "FOCUS_PROTECTION",
        "FOCUS_RESEARCH",
        "FRUIT_SPECIAL",
        "GAIA_SPECIAL",
        "PLC_RACIAL_PURITY",
        "PLC_TECHNOCRACY",
        "POSITRONIUM_SPECIAL",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *AVERAGE_RESEARCH,
        *GOOD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *GOOD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_HHHOH"),
        *AVERAGE_SUPPLY,
        *GREAT_DEFENSE_TROOPS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=TUNDRA_STANDARD_EP,
    graphic="icons/species/hhhoh.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_SPACE_FLUX_BASICS",
    description="SHP_SPACE_FLUX_BASICS_DESC",
    short_description="SHIP_WEAPON_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=10 * TECH_COST_MULTIPLIER,
    researchturns=2,
    tags=["PEDIA_SR_WEAPON_TECHS"],
    prerequisites=["SHP_ROOT_AGGRESSION"],
    unlock=Item(type=UnlockShipPart, name="SR_FLUX_LANCE"),
    graphic="icons/ship_parts/flux-lance.png",
)



from collections import OrderedDict
from collections.abc import Iterable

from common.fo_typing import PlanetId, SpeciesName
from freeorion_tools.caching import cache_for_current_turn

_the_planner = None


@cache_for_current_turn
def colonies_targeted() -> frozenset[PlanetId]:
    return _the_planner.colonies_targeted()


@cache_for_current_turn
def outposts_targeted() -> frozenset[PlanetId]:
    return _the_planner.outposts_targeted()


def get_colonisable_planet_ids(include_targeted: bool = False) -> OrderedDict[PlanetId, tuple[float, SpeciesName]]:
    return _the_planner.get_colonisable_planet_ids(include_targeted)


def get_colonisable_outpost_ids(include_targeted: bool = False) -> OrderedDict[PlanetId, tuple[float, SpeciesName]]:
    return _the_planner.get_colonisable_outpost_ids(include_targeted)


def update_colonisable_planet_ids(new_list: Iterable) -> None:
    return _the_planner.update_colonisable_planet_ids(new_list)


def update_colonisable_outpost_ids(new_list: Iterable) -> None:
    return _the_planner.update_colonisable_outpost_ids(new_list)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_HEAVY_AST_HULL",
    description="SHP_HEAVY_AST_HULL_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=128 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_ASTEROID_HULL_TECHS"],
    prerequisites=["SHP_ASTEROID_REFORM"],
    unlock=[Item(type=UnlockShipHull, name="SH_HEAVY_ASTEROID")],
    graphic="icons/ship_hulls/heavy_asteroid_hull_small.png",
)



from CombatRatingsAI._fleet_combat_stats import get_ships_stats_for_fleet
from CombatRatingsAI._ratings import (
    default_ship_stats,
    get_empire_standard_military_ship_stats,
    get_fleet_rating,
    get_fleet_rating_against_planets,
    get_ship_rating,
    rating_difference,
    rating_needed,
    species_shield_bonus,
)
from CombatRatingsAI._ship_combat_stats import ShipCombatStats, get_ship_combat_stats
from CombatRatingsAI._targets import get_allowed_targets



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import TERRAN_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import GOOD_HAPPINESS
from species.species_macros.industry import NO_INDUSTRY
from species.species_macros.influence import VERY_BAD_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import GREAT_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS
from species.species_macros.weapons import BAD_WEAPONS

Species(
    name="SP_SILEXIAN",
    description="SP_SILEXIAN_DESC",
    gameplay_description="SP_SILEXIAN_GAMEPLAY_DESC",
    native=True,
    can_colonize=True,
    tags=[
        "ROBOTIC",
        "ARTISTIC",
        "NO_INDUSTRY",
        "BAD_RESEARCH",
        "VERY_BAD_INFLUENCE",
        "GOOD_HAPPINESS",
        "GREAT_SUPPLY",
        "BAD_WEAPONS",
        "PEDIA_ROBOTIC_SPECIES_CLASS",
        "PEDIA_ARTISTIC",
    ],
    foci=[
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "MIMETIC_ALLOY_SPECIAL",
        "SHIMMER_SILK_SPECIAL",
        "MONOPOLE_SPECIAL",
        "COMPUTRONIUM_SPECIAL",
        "PHILOSOPHER_SPECIAL",
        "TEMPORAL_ANOMALY_SPECIAL",
        "PLC_DIVINE_AUTHORITY",
        "PLC_CENTRALIZATION",
        "PLC_MARTIAL_LAW",
        "PLC_ALGORITHMIC_RESEARCH",
        "PLC_INDUSTRIALISM",
        "PLC_ENGINEERING",
        "PLC_BUREAUCRACY",
    ],
    dislikes=[
        "BLD_AUTO_HISTORY_ANALYSER",
        "BLD_IMPERIAL_PALACE",
        "BLD_REGIONAL_ADMIN",
        "BLD_MILITARY_COMMAND",
        "BLD_BIOTERROR_PROJECTOR",
        "BLD_LIGHTHOUSE",
        "BLD_STARGATE",
        "MINERALS_SPECIAL",
        "FRUIT_SPECIAL",
        "ELERIUM_SPECIAL",
        "KRAKEN_IN_THE_ICE_SPECIAL",
        "PLC_LIBERTY",
        "PLC_CONFEDERATION",
        "PLC_ARTISAN_WORKSHOPS",
        "PLC_DESIGN_SIMPLICITY",
        "PLC_EXPLORATION",
    ],
    effectsgroups=[
        NO_INDUSTRY,
        *BAD_RESEARCH,
        *VERY_BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *GOOD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_SILEXIAN"),
        *GREAT_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        *BAD_WEAPONS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=TERRAN_STANDARD_EP,
    graphic="icons/species/robotic-06.png",
)



import freeOrionAIInterface as fo
from logging import debug, warning
from typing import Optional

import AIstate
import fleet_orders
import pathfinding
import PlanetUtilsAI
from AIDependencies import DRYDOCK_HAPPINESS_THRESHOLD, INVALID_ID
from aistate_interface import get_aistate
from buildings import get_empire_drydocks
from common.fo_typing import SystemId
from freeorion_tools import get_fleet_position
from target import TargetFleet, TargetSystem
from turn_state import get_system_supply


def create_move_orders_to_system(fleet: TargetFleet, target: TargetSystem) -> list["fleet_orders.OrderMove"]:
    """
    Create a list of move orders from the fleet's current system to the target system.

    :param fleet: Fleet to be moved
    :param target: target system
    :return: list of move orders
    """
    # TODO: use Graph Theory to construct move orders
    # TODO: add priority
    starting_system = fleet.get_system()  # current fleet location or current target system if on starlane
    if starting_system == target:
        # nothing to do here
        return []
    # if the mission does not end at the targeted system, make sure we can actually return to supply after moving.
    ensure_return = target.id not in set(
        AIstate.colonyTargetedSystemIDs + AIstate.outpostTargetedSystemIDs + AIstate.invasionTargetedSystemIDs
    )
    system_targets = can_travel_to_system(fleet.id, starting_system, target, ensure_return=ensure_return)
    result = [fleet_orders.OrderMove(fleet, system) for system in system_targets]
    if not result and starting_system.id != target.id:
        warning(f"fleet {fleet.id} can't travel to system {target}")
    return result


def can_travel_to_system(
    fleet_id: int, start: TargetSystem, target: TargetSystem, ensure_return: bool = False
) -> list[TargetSystem]:
    """
    Return list systems to be visited.
    """
    if start == target:
        return [TargetSystem(start.id)]

    debug(f"Requesting path for fleet {fo.getUniverse().getFleet(fleet_id)} from {start} to {target}")
    target_distance_from_supply = -min(get_system_supply(target.id), 0)

    # low-aggression AIs may not travel far from supply
    if not get_aistate().character.may_travel_beyond_supply(target_distance_from_supply):
        debug("May not move %d out of supply" % target_distance_from_supply)
        return []

    min_fuel_at_target = target_distance_from_supply if ensure_return else 0
    path_info = pathfinding.find_path_with_resupply(
        start.id, target.id, fleet_id, minimum_fuel_at_target=min_fuel_at_target
    )
    if path_info is None:
        debug("Found no valid path.")
        return []

    debug("Found valid path: %s" % str(path_info))
    return [TargetSystem(sys_id) for sys_id in path_info.path]


def get_nearest_supplied_system(start_system_id: SystemId):
    """Return systemAITarget of nearest supplied system from starting system startSystemID."""
    empire = fo.getEmpire()
    fleet_supplyable_system_ids = empire.fleetSupplyableSystemIDs
    universe = fo.getUniverse()

    if start_system_id in fleet_supplyable_system_ids:
        return TargetSystem(start_system_id)
    else:
        min_jumps = 9999  # infinity
        supply_system_id = INVALID_ID
        for system_id in fleet_supplyable_system_ids:
            if start_system_id != INVALID_ID and system_id != INVALID_ID:
                least_jumps_len = universe.jumpDistance(start_system_id, system_id)
                if least_jumps_len < min_jumps:
                    min_jumps = least_jumps_len
                    supply_system_id = system_id
        return TargetSystem(supply_system_id)


def get_best_drydock_system_id(start_system_id: int, fleet_id: int) -> Optional[int]:
    """
    Get system_id of best drydock capable of repair, where best is nearest drydock
    that has a current and target happiness greater than the HAPPINESS_THRESHOLD
    with a path that is not blockaded or that the fleet can fight through to with
    acceptable losses.

    :param start_system_id: current location of fleet - used to find closest target
    :param fleet_id: fleet that needs path to drydock
    :return: most suitable system id where the fleet should be repaired.
    """
    if start_system_id == INVALID_ID:
        warning("get_best_drydock_system_id passed bad system id.")
        return None

    if fleet_id == INVALID_ID:
        warning("get_best_drydock_system_id passed bad fleet id.")
        return None

    universe = fo.getUniverse()
    start_system = TargetSystem(start_system_id)
    drydock_system_ids = set()
    for sys_id, pids in get_empire_drydocks().items():
        if sys_id == INVALID_ID:
            warning("get_best_drydock_system_id passed bad drydock sys_id.")
            continue
        for pid in pids:
            planet = universe.getPlanet(pid)
            if (
                planet
                and planet.currentMeterValue(fo.meterType.happiness) >= DRYDOCK_HAPPINESS_THRESHOLD
                and planet.currentMeterValue(fo.meterType.targetHappiness) >= DRYDOCK_HAPPINESS_THRESHOLD
            ):
                drydock_system_ids.add(sys_id)
                break

    sys_distances = sorted([(universe.jumpDistance(start_system_id, sys_id), sys_id) for sys_id in drydock_system_ids])

    aistate = get_aistate()
    fleet_rating = aistate.get_rating(fleet_id)
    for _, dock_sys_id in sys_distances:
        dock_system = TargetSystem(dock_sys_id)
        path = can_travel_to_system(fleet_id, start_system, dock_system)

        path_rating = sum([aistate.systemStatus[path_sys.id]["totalThreat"] for path_sys in path])

        SAFETY_MARGIN = 10
        if SAFETY_MARGIN * path_rating <= fleet_rating:
            debug(
                f"Drydock recommendation {dock_system} from {start_system} for fleet {universe.getFleet(fleet_id)} with fleet rating {fleet_rating:.1f} and path rating {path_rating:.1f}."
            )
            return dock_system.id

    debug(
        f"No safe drydock recommendation from {start_system} for fleet {universe.getFleet(fleet_id)} with fleet rating {fleet_rating:.1f}."
    )
    return None


def get_safe_path_leg_to_dest(fleet_id, start_id, dest_id):
    start_targ = TargetSystem(start_id)
    dest_targ = TargetSystem(dest_id)
    # TODO actually get a safe path
    this_path = can_travel_to_system(fleet_id, start_targ, dest_targ, ensure_return=False)
    path_ids = [targ.id for targ in this_path if targ.id != start_id] + [start_id]
    universe = fo.getUniverse()
    debug(
        "Fleet %d requested safe path leg from %s to %s, found path %s"
        % (fleet_id, universe.getSystem(start_id), universe.getSystem(dest_id), PlanetUtilsAI.sys_name_ids(path_ids))
    )
    return path_ids[0]


def get_resupply_fleet_order(fleet_target: TargetFleet) -> "fleet_orders.OrderResupply":
    """
    Return fleet_orders.OrderResupply to nearest supplied system.
    """
    # find nearest supplied system
    supplied_system_target = get_nearest_supplied_system(get_fleet_position(fleet_target.id))
    # create resupply AIFleetOrder
    return fleet_orders.OrderResupply(fleet_target, supplied_system_target)


def get_repair_fleet_order(fleet: TargetFleet) -> Optional["fleet_orders.OrderRepair"]:
    """
    Return fleet_orders.OrderRepair for fleet to proceed to system with drydock.
    """
    # TODO Cover new mechanics where happiness increases repair rate - don't always use nearest system!
    # find nearest drydock system
    drydock_sys_id = get_best_drydock_system_id(get_fleet_position(fleet.id), fleet.id)
    if drydock_sys_id is None:
        return None

    debug(f"Ordering fleet {fleet} to {fo.getUniverse().getSystem(drydock_sys_id)} for repair")
    return fleet_orders.OrderRepair(fleet, TargetSystem(drydock_sys_id))



from focs._effects import (
    EffectsGroup,
    InSystem,
    LocalCandidate,
    OwnedBy,
    SetStructure,
    Ship,
    Source,
    Structure,
    Turn,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.misc import SHIP_STRUCTURE_FACTOR

Tech(
    name="SHP_BASIC_DAM_CONT",
    description="SHP_BASIC_DAM_CONT_DESC",
    short_description="STRUCTURE_SHORT_DESC",
    category="SHIP_PARTS_CATEGORY",
    researchcost=80 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_DAMAGE_CONTROL_PART_TECHS"],
    prerequisites=["SHP_MIL_ROBO_CONT"],
    unlock=Item(type=UnlockPolicy, name="PLC_ENGINEERING"),
    effectsgroups=[
        EffectsGroup(
            scope=Ship
            & OwnedBy(empire=Source.Owner)
            & (~InSystem() | InSystem() & Turn(low=LocalCandidate.System.LastTurnBattleHere + 1))
            & Structure(high=LocalCandidate.MaxStructure - 0.001),
            effects=SetStructure(value=Value + SHIP_STRUCTURE_FACTOR),
        )
    ],
)



from collections.abc import Mapping

from common.fo_typing import PlanetId, SpecialName
from empire.survey_lock import survey_universe_lock
from freeorion_tools.caching import cache_for_current_turn


def set_growth_special(special: SpecialName, pid: PlanetId):
    _get_growth_specials().setdefault(special, []).append(pid)


@survey_universe_lock
def get_growth_specials() -> Mapping[SpecialName, list[PlanetId]]:
    """
    Return map from a growth special to list of the planets where we have it.
    """
    return _get_growth_specials()


@cache_for_current_turn
def _get_growth_specials() -> dict[SpecialName, list[PlanetId]]:
    """
    Return mutable state.
    """
    return {}



from focs._effects import (
    EffectsGroup,
    NamedRealLookup,
    OwnedBy,
    SetMaxStructure,
    Ship,
    Source,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_REINFORCED_HULL",
    description="SHP_REINFORCED_HULL_DESC",
    short_description="STRUCTURE_SHORT_DESC",
    category="SHIP_PARTS_CATEGORY",
    researchcost=72 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_DAMAGE_CONTROL_PART_TECHS"],
    prerequisites=["CON_ARCH_MONOFILS"],
    effectsgroups=[
        EffectsGroup(
            scope=Ship & OwnedBy(empire=Source.Owner),
            effects=SetMaxStructure(value=Value + NamedRealLookup(name="SHP_REINFORCED_HULL_BONUS")),
        )
    ],
    graphic="icons/tech/structural_integrity_fields.png",
)



from focs._species import *
from species.species_macros.detection import BAD_DETECTION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import DESERT_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import BAD_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import GOOD_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import GREAT_SUPPLY
from species.species_macros.troops import GOOD_DEFENSE_TROOPS
from species.species_macros.weapons import GOOD_WEAPONS

Species(
    name="SP_ETTY",
    description="SP_ETTY_DESC",
    gameplay_description="SP_ETTY_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ROBOTIC",
        "BAD_INDUSTRY",
        "GOOD_WEAPONS",
        "GREAT_SUPPLY",
        "BAD_DETECTION",
        "GOOD_STEALTH",
        "PEDIA_ROBOTIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "FOCUS_RESEARCH",
        "MIMETIC_ALLOY_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "SUPERCONDUCTOR_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "COMPUTRONIUM_SPECIAL",
        "FORTRESS_SPECIAL",
        "PLC_ALLIED_REPAIR",
        "PLC_CENTRALIZATION",
        "PLC_LIBERTY",
        "PLC_CONTINUOUS_SCANNING",
        "PLC_BUREAUCRACY",
    ],
    dislikes=[
        "BLD_SHIPYARD_AST",
        "BLD_SHIPYARD_AST_REF",
        "BLD_SPATIAL_DISTORT_GEN",
        "BLD_SPACE_ELEVATOR",
        "BLD_STOCKPILING_CENTER",
        "BLD_LIGHTHOUSE",
        "FRUIT_SPECIAL",
        "PANOPTICON_SPECIAL",
        "TEMPORAL_ANOMALY_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "PLC_CONFEDERATION",
        "PLC_NATIVE_APPROPRIATION",
        "PLC_INDUSTRIALISM",
        "PLC_EXOBOT_PRODUCTIVITY",
    ],
    effectsgroups=[
        *BAD_INDUSTRY,
        *AVERAGE_RESEARCH,
        *AVERAGE_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_ETTY"),
        *GREAT_SUPPLY,
        *GOOD_DEFENSE_TROOPS,
        *GOOD_WEAPONS,
        *BAD_DETECTION,
        *GOOD_STEALTH,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=DESERT_STANDARD_EP,
    graphic="icons/species/etty.png",
)



import freeOrionAIInterface as fo
from logging import debug, error, warning

import CombatRatingsAI
import EspionageAI
import ExplorationAI
import FleetUtilsAI
import MilitaryAI
import MoveUtilsAI
from aistate_interface import get_aistate
from EnumsAI import MissionType, ShipRoleType
from freeorion_tools import combine_ratings
from target import Target, TargetFleet, TargetPlanet, TargetSystem


class AIFleetOrder:
    """Stores information about orders which can be executed."""

    TARGET_TYPE = None
    ORDER_NAME = ""

    def __init__(self, fleet: TargetFleet, target: Target):
        """
        :param fleet: fleet to execute order
        :param target: fleet target, depends of order type
        """
        if not isinstance(fleet, TargetFleet):
            error("Order required fleet got %s" % type(fleet))

        if not isinstance(target, self.TARGET_TYPE):
            error(f"Target is not allowed, got {type(target)} expect {self.TARGET_TYPE}")

        self.fleet = fleet
        self.target = target
        self.executed = False
        self.order_issued = False

    def __setstate__(self, state):
        # construct the universe objects from stored ids
        state["fleet"] = TargetFleet(state["fleet"])
        target_type = state.pop("target_type")
        if state["target"] is not None:
            assert self.TARGET_TYPE.object_name == target_type
            state["target"] = self.TARGET_TYPE(state["target"])
        self.__dict__ = state

    def __getstate__(self):
        retval = dict(self.__dict__)
        # do not store the universe object but only the fleet id
        retval["fleet"] = self.fleet.id
        if self.target is not None:
            retval["target"] = self.target.id
            retval["target_type"] = self.target.object_name
        else:
            retval["target_type"] = None
        return retval

    def is_valid(self):
        """Check if FleetOrder could be somehow in future issued = is valid."""
        if self.executed and self.order_issued:
            debug("\t\t order not valid because already executed and completed")
            return False
        if self.fleet and self.target:
            return True
        else:
            debug(f"\t\t order not valid: fleet validity: {bool(self.fleet)} and target validity {bool(self.target)}")
            return False

    def can_issue_order(self, verbose=False):
        """If FleetOrder can be issued now."""
        # for some orders, may need to re-issue if invasion/outposting/colonization was interrupted
        if self.executed and not isinstance(self, (OrderOutpost, OrderColonize, OrderInvade)):
            return False
        if not self.is_valid():
            return False

        if verbose:
            sys1 = self.fleet.get_system()
            main_fleet_mission = get_aistate().get_fleet_mission(self.fleet.id)
            debug(
                "  Can issue %s - Mission Type %s (%s), current loc sys %d - %s"
                % (self, main_fleet_mission.type, main_fleet_mission.type, self.fleet.id, sys1)
            )
        return True

    def issue_order(self):
        if not self.can_issue_order():  # appears to be redundant with check in IAFleetMission?
            debug("  can't issue %s" % self)
            return False
        # by default we now set the order as issue and executed.  For any subclass where order issuence and execution
        # is not necessarily sure, these values can be reset after any appropriate checks in the respective
        # subclass issue_order()
        self.order_issued = True
        self.executed = True
        return True

    def __str__(self):
        execute_status = "in progress"
        if self.executed:
            execute_status = "executed"
        elif self.order_issued:
            execute_status = "order issued"
        return f"[{self.ORDER_NAME}] of {self.fleet.get_object()} to {self.target.get_object()} {execute_status}"

    def __eq__(self, other):
        return type(self) is type(other) and self.fleet == other.fleet and self.target == other.target

    def __hash__(self):
        return hash(self.fleet)


class OrderMove(AIFleetOrder):
    ORDER_NAME = "move"
    TARGET_TYPE = TargetSystem

    def can_issue_order(self, verbose=False):  # noqa: C901
        if not super().can_issue_order(verbose=verbose):
            return False
        # TODO: figure out better way to have invasions (& possibly colonizations)
        #       require visibility on target without needing visibility of all intermediate systems
        # if False and main_mission_type not in [MissionType.ATTACK,  # TODO: consider this later
        #                                        MissionType.MILITARY,
        #                                        MissionType.SECURE,
        #                                        MissionType.HIT_AND_RUN,
        #                                        MissionType.EXPLORATION]:
        #     if not universe.getVisibility(target_id, get_aistate().empireID) >= fo.visibility.partial:
        #         #if not target_id in interior systems
        #         get_aistate().needsEmergencyExploration.append(fleet.systemID)
        #         return False
        system_id = self.fleet.get_system().id
        if system_id == self.target.get_system().id:
            return True  # TODO: already there, but could consider retreating

        aistate = get_aistate()
        main_fleet_mission = aistate.get_fleet_mission(self.fleet.id)

        # TODO: Rate against specific enemies here
        fleet_rating = CombatRatingsAI.get_fleet_rating(self.fleet.id)
        fleet_rating_vs_planets = CombatRatingsAI.get_fleet_rating_against_planets(self.fleet.id)
        target_sys_status = aistate.systemStatus.get(self.target.id, {})
        f_threat = target_sys_status.get("fleetThreat", 0)
        m_threat = target_sys_status.get("monsterThreat", 0)
        p_threat = target_sys_status.get("planetThreat", 0)
        threat = f_threat + m_threat + p_threat
        safety_factor = aistate.character.military_safety_factor()
        universe = fo.getUniverse()
        if main_fleet_mission.type == MissionType.INVASION and not trooper_move_reqs_met(
            main_fleet_mission.target, self, verbose
        ):
            return False
        if fleet_rating >= safety_factor * threat and fleet_rating_vs_planets >= p_threat:
            return True
        elif not p_threat and self.target.id in fo.getEmpire().supplyUnobstructedSystems:
            return True
        else:
            sys1 = universe.getSystem(system_id)
            sys1_name = sys1 and sys1.name or "unknown"
            target_system = self.target.get_system()
            target_system_name = (target_system and target_system.get_object().name) or "unknown"
            # TODO: adjust calc for any departing fleets
            my_other_fleet_rating = aistate.systemStatus.get(self.target.id, {}).get("myFleetRating", 0)
            my_other_fleet_rating_vs_planets = aistate.systemStatus.get(self.target.id, {}).get(
                "myFleetRatingVsPlanets", 0
            )
            is_military = aistate.get_fleet_role(self.fleet.id) == MissionType.MILITARY

            total_rating = combine_ratings(my_other_fleet_rating, fleet_rating)
            total_rating_vs_planets = combine_ratings(my_other_fleet_rating_vs_planets, fleet_rating_vs_planets)
            if my_other_fleet_rating > 3 * safety_factor * threat or (
                is_military and total_rating_vs_planets > 2.5 * p_threat and total_rating > safety_factor * threat
            ):
                debug(
                    "\tAdvancing fleet %d (rating %d) at system %d (%s) into system %d (%s) with threat %d"
                    " because of sufficient empire fleet strength already at destination"
                    % (
                        self.fleet.id,
                        fleet_rating,
                        system_id,
                        sys1_name,
                        self.target.id,
                        target_system_name,
                        threat,
                    )
                )
                return True
            elif (
                threat == p_threat
                and not self.fleet.get_object().aggressive
                and not my_other_fleet_rating
                and not target_sys_status.get("localEnemyFleetIDs", [-1])
            ):
                if verbose:
                    debug(
                        "\tAdvancing fleet %d (rating %d) at system %d (%s) "
                        "into system %d (%s) with planet threat %d because non aggressive"
                        " and no other fleets present to trigger combat"
                        % (
                            self.fleet.id,
                            fleet_rating,
                            system_id,
                            sys1_name,
                            self.target.id,
                            target_system_name,
                            threat,
                        )
                    )
                return True
            else:
                if verbose:
                    _info = (
                        self.fleet.id,
                        fleet_rating,
                        system_id,
                        sys1_name,
                        self.target.id,
                        target_system_name,
                        threat,
                    )
                    debug(
                        "\tHolding fleet %d (rating %d) at system %d (%s) "
                        "before travelling to system %d (%s) with threat %d" % _info
                    )
                needs_vis = aistate.misc.setdefault("needs_vis", [])
                if self.target.id not in needs_vis:
                    needs_vis.append(self.target.id)
                return False

    def issue_order(self):
        if not super().issue_order():
            return False
        fleet_id = self.fleet.id
        system_id = self.target.get_system().id
        fleet = self.fleet.get_object()
        if system_id not in [fleet.systemID, fleet.nextSystemID]:
            dest_id = system_id
            fo.issueFleetMoveOrder(fleet_id, dest_id)
            debug(f"Order issued: {self.ORDER_NAME} fleet: {self.fleet} target: {self.target}")
        aistate = get_aistate()
        if system_id == fleet.systemID:
            if aistate.get_fleet_role(fleet_id) == MissionType.EXPLORATION:
                if system_id in aistate.needsEmergencyExploration:
                    aistate.needsEmergencyExploration.remove(system_id)
        return True


class OrderPause(AIFleetOrder):
    """Ensure Fleet at least temporarily halts movement at the target system."""

    ORDER_NAME = "pause"
    TARGET_TYPE = TargetSystem

    def is_valid(self):
        if not super().is_valid():
            return False
        return bool(self.target.get_system().get_object())

    def issue_order(self):
        if not super().issue_order():
            return False
        # not executed until actually arrives at target system
        self.executed = self.fleet.get_current_system_id() == self.target.get_system().id


class OrderResupply(AIFleetOrder):
    ORDER_NAME = "resupply"
    TARGET_TYPE = TargetSystem

    def is_valid(self):
        if not super().is_valid():
            return False
        return self.target.id in fo.getEmpire().fleetSupplyableSystemIDs

    def issue_order(self):
        if not super().issue_order():
            return False
        fleet_id = self.fleet.id
        system_id = self.target.get_system().id
        fleet = self.fleet.get_object()
        aistate = get_aistate()
        if system_id == fleet.systemID:
            if aistate.get_fleet_role(fleet_id) == MissionType.EXPLORATION:
                if system_id in aistate.needsEmergencyExploration:
                    aistate.needsEmergencyExploration.remove(system_id)
            return True
        if system_id != fleet.nextSystemID:
            self.executed = False
            start_id = FleetUtilsAI.get_fleet_system(fleet)
            dest_id = MoveUtilsAI.get_safe_path_leg_to_dest(fleet_id, start_id, system_id)
            universe = fo.getUniverse()
            debug(
                "fleet %d with order type(%s) sent to safe leg dest %s and ultimate dest %s"
                % (fleet_id, self.ORDER_NAME, universe.getSystem(dest_id), universe.getSystem(system_id))
            )
            fo.issueFleetMoveOrder(fleet_id, dest_id)
            debug(f"Order issued: {self.ORDER_NAME} fleet: {self.fleet} target: {self.target}")
        return True


class OrderOutpost(AIFleetOrder):
    ORDER_NAME = "outpost"
    TARGET_TYPE = TargetPlanet

    def is_valid(self):
        if not super().is_valid():
            return False
        planet = self.target.get_object()
        if not planet.unowned:
            # terminate early
            self.executed = True
            self.order_issued = True
            return False
        else:
            return self.fleet.get_object().hasOutpostShips

    def can_issue_order(self, verbose=False):
        # TODO: check for separate fleet holding outpost ships
        if not super().can_issue_order(verbose=verbose):
            return False
        universe = fo.getUniverse()
        ship_id = FleetUtilsAI.get_ship_id_with_role(self.fleet.id, ShipRoleType.CIVILIAN_OUTPOST)
        if ship_id is None:
            ship_id = FleetUtilsAI.get_ship_id_with_role(self.fleet.id, ShipRoleType.BASE_OUTPOST)
        ship = universe.getShip(ship_id)
        return ship is not None and self.fleet.get_object().systemID == self.target.get_system().id and ship.canColonize

    def issue_order(self):
        if not super().issue_order():
            return False
        # we can't know yet if the order will actually execute; instead, rely on the fact that if the order does get
        # executed, then next turn it will be invalid
        self.executed = False
        planet = self.target.get_object()
        if not planet.unowned:
            return False
        fleet_id = self.fleet.id
        ship_id = FleetUtilsAI.get_ship_id_with_role(fleet_id, ShipRoleType.CIVILIAN_OUTPOST)
        if ship_id is None:
            ship_id = FleetUtilsAI.get_ship_id_with_role(fleet_id, ShipRoleType.BASE_OUTPOST)
        if fo.issueColonizeOrder(ship_id, self.target.id):
            debug(f"Order issued: {self.ORDER_NAME} fleet: {self.fleet} target: {self.target}")
            return True
        else:
            self.order_issued = False
            warning(f"Order issuance failed: {self.ORDER_NAME} fleet: {self.fleet} target: {self.target}")
            return False


class OrderColonize(AIFleetOrder):
    ORDER_NAME = "colonize"
    TARGET_TYPE = TargetPlanet

    def issue_order(self):
        if not super().issue_order():
            return False
        # we can't know yet if the order will actually execute; instead, rely on the fact that if the order does get
        # executed, then next turn it will be invalid
        self.executed = False

        fleet_id = self.fleet.id
        ship_id = FleetUtilsAI.get_ship_id_with_role(fleet_id, ShipRoleType.CIVILIAN_COLONISATION)
        if ship_id is None:
            ship_id = FleetUtilsAI.get_ship_id_with_role(fleet_id, ShipRoleType.BASE_COLONISATION)

        if fo.issueColonizeOrder(ship_id, self.target.id):
            debug(f"Order issued: {self.ORDER_NAME} fleet: {self.fleet} target: {self.target}")
            return True
        else:
            self.order_issued = False
            warning(f"Order issuance failed: {self.ORDER_NAME} fleet: {self.fleet} target: {self.target}")
            return False

    def is_valid(self):
        if not super().is_valid():
            return False
        planet = self.target.get_object()
        if (planet.unowned or planet.ownedBy(fo.empireID())) and not planet.currentMeterValue(fo.meterType.population):
            return self.fleet.get_object().hasColonyShips
        # Otherwise, terminate early
        self.executed = True
        self.order_issued = True
        return False

    def can_issue_order(self, verbose=False):
        if not super().is_valid():
            return False
        # TODO: check for separate fleet holding colony ships
        ship_id = FleetUtilsAI.get_ship_id_with_role(self.fleet.id, ShipRoleType.CIVILIAN_COLONISATION)
        if ship_id is None:
            ship_id = FleetUtilsAI.get_ship_id_with_role(self.fleet.id, ShipRoleType.BASE_COLONISATION)
        universe = fo.getUniverse()
        ship = universe.getShip(ship_id)
        if ship and not ship.canColonize:
            warning("colonization fleet %d has no colony ship" % self.fleet.id)
        return ship is not None and self.fleet.get_object().systemID == self.target.get_system().id and ship.canColonize


class OrderDefend(AIFleetOrder):
    """
    Used for orbital defense, have no real orders.
    """

    ORDER_NAME = "defend"
    TARGET_TYPE = TargetSystem


class OrderInvade(AIFleetOrder):
    ORDER_NAME = "invade"
    TARGET_TYPE = TargetPlanet

    def is_valid(self):
        if not super().is_valid():
            return False
        planet = self.target.get_object()
        planet_population = planet.currentMeterValue(fo.meterType.population)
        if planet.unowned and not planet_population:
            debug(
                f"\t\t invasion order not valid due to target planet status-- owned: {not planet.unowned} and population {planet_population:.1f}"
            )
            # terminate early
            self.executed = True
            self.order_issued = True
            return False
        else:
            return self.fleet.get_object().hasTroopShips

    def can_issue_order(self, verbose=False):
        if not super().is_valid():
            return False
        # TODO: check for separate fleet holding invasion ships
        ship_id = FleetUtilsAI.get_ship_id_with_role(self.fleet.id, ShipRoleType.MILITARY_INVASION, False)
        if ship_id is None:
            ship_id = FleetUtilsAI.get_ship_id_with_role(self.fleet.id, ShipRoleType.BASE_INVASION)
        universe = fo.getUniverse()
        ship = universe.getShip(ship_id)
        planet = self.target.get_object()
        return all(
            (
                ship is not None,
                self.fleet.get_object().systemID == planet.systemID,
                ship.canInvade,
                not planet.initialMeterValue(fo.meterType.shield),
            )
        )

    def issue_order(self):
        if not super().can_issue_order():
            return False

        universe = fo.getUniverse()
        planet_id = self.target.id
        planet = self.target.get_object()
        fleet = self.fleet.get_object()

        invasion_roles = (ShipRoleType.BASE_INVASION, ShipRoleType.MILITARY_INVASION)

        debug(f"Issuing order: {self.ORDER_NAME} fleet: {self.fleet} target: {self.target}")
        # will track if at least one invasion troops successfully deployed
        result = True
        aistate = get_aistate()
        overkill_margin = 2  # TODO: get from character module; allows a handful of extra troops to be immediately
        #                            defending planet
        # invasion orders processed before regen takes place, so use initialMeterValue() here
        troops_wanted = planet.initialMeterValue(fo.meterType.troops) + overkill_margin
        troops_already_assigned = 0  # TODO: get from other fleets in same system
        troops_assigned = 0
        # Todo: evaluate all local troop ships (including other fleets) before using any, make sure base invasion troops
        #       are used first, and that not too many altogether are used (choosing an optimal collection to use).
        for invasion_role in invasion_roles:  # first checks base troops, then regular
            if not result:
                break
            for ship_id in fleet.shipIDs:
                if troops_already_assigned + troops_assigned >= troops_wanted:
                    break
                ship = universe.getShip(ship_id)
                if aistate.get_ship_role(ship.design.id) != invasion_role:
                    continue

                debug("Ordering troop ship %d to invade %s" % (ship_id, planet))
                result = fo.issueInvadeOrder(ship_id, planet_id) and result
                if not result:
                    shields = planet.currentMeterValue(fo.meterType.shield)
                    planet_stealth = planet.currentMeterValue(fo.meterType.stealth)
                    pop = planet.currentMeterValue(fo.meterType.population)
                    warning("Invasion order failed!")
                    debug(
                        " -- planet has %.1f stealth, shields %.1f, %.1f population and "
                        "is owned by empire %d" % (planet_stealth, shields, pop, planet.owner)
                    )
                    if "needsEmergencyExploration" not in dir(aistate):
                        aistate.needsEmergencyExploration = []

                    # TODO: Check if this even makes sense - to invade a planet the ship must be in the system
                    #       which should grant the same visibility as a scout ship...
                    ExplorationAI.request_emergency_exploration(fleet.systemID)
                    debug("Due to trouble invading, added system %d to Emergency Exploration List" % fleet.systemID)
                    self.executed = False
                    # debug(universe.getPlanet(planet_id).dump())  # TODO: fix fo.UniverseObject.dump()
                    break
                troops_assigned += ship.troopCapacity
        # TODO: split off unused troop ships into new fleet and give new orders this cycle
        if result:
            debug("Successfully ordered %d troopers to invade %s" % (troops_assigned, planet))
            return True
        else:
            return False


class OrderMilitary(AIFleetOrder):
    ORDER_NAME = "military"
    TARGET_TYPE = TargetSystem

    def is_valid(self):
        if not super().is_valid():
            return False
        fleet = self.fleet.get_object()
        # TODO: consider bombardment-only fleets/orders
        return fleet is not None and (fleet.hasArmedShips or fleet.hasFighterShips)

    def can_issue_order(self, verbose=False):
        # TODO: consider bombardment
        # TODO: consider simmply looking at fleet characteristics, as is done for is_valid()
        ship_id = FleetUtilsAI.get_ship_id_with_role(self.fleet.id, ShipRoleType.MILITARY)
        universe = fo.getUniverse()
        ship = universe.getShip(ship_id)
        return (
            ship is not None
            and self.fleet.get_object().systemID == self.target.id
            and (ship.isArmed or ship.hasFighters)
        )

    def issue_order(self):
        if not super().issue_order():
            return False
        target_sys_id = self.target.id
        fleet = self.target.get_object()
        system_status = get_aistate().systemStatus.get(target_sys_id, {})
        total_threat = sum(system_status.get(threat, 0) for threat in ("fleetThreat", "planetThreat", "monsterThreat"))
        combat_trigger = system_status.get("fleetThreat", 0) or system_status.get("monsterThreat", 0)
        if not combat_trigger and system_status.get("planetThreat", 0):
            universe = fo.getUniverse()
            system = universe.getSystem(target_sys_id)
            for planet_id in system.planetIDs:
                planet = universe.getPlanet(planet_id)
                if planet.ownedBy(fo.empireID()):  # TODO: also exclude at-peace planets
                    continue
                if planet.unowned and not EspionageAI.colony_detectable_by_empire(planet_id, empire=fo.empireID()):
                    continue
                if sum(
                    [
                        planet.currentMeterValue(meter_type)
                        for meter_type in [fo.meterType.defense, fo.meterType.shield, fo.meterType.construction]
                    ]
                ):
                    combat_trigger = True
                    break
        if not all(
            (
                fleet,
                fleet.systemID == target_sys_id,
                system_status.get("currently_visible", False),
                not (total_threat and combat_trigger),
            )
        ):
            self.executed = False
        return True


class OrderRepair(AIFleetOrder):
    ORDER_NAME = "repair"
    TARGET_TYPE = TargetSystem

    def is_valid(self):
        if not super().is_valid():
            return False
        return self.target.id in fo.getEmpire().fleetSupplyableSystemIDs  # TODO: check for drydock still there/owned

    def issue_order(self):
        if not super().issue_order():
            return False
        fleet_id = self.fleet.id
        system_id = self.target.get_system().id
        fleet = self.fleet.get_object()  # type: fo.fleet
        if system_id == fleet.systemID:
            aistate = get_aistate()
            if aistate.get_fleet_role(fleet_id) == MissionType.EXPLORATION:
                if system_id in aistate.needsEmergencyExploration:
                    aistate.needsEmergencyExploration.remove(system_id)
        elif system_id != fleet.nextSystemID:
            fo.issueAggressionOrder(fleet_id, False)
            start_id = FleetUtilsAI.get_fleet_system(fleet)
            dest_id = MoveUtilsAI.get_safe_path_leg_to_dest(fleet_id, start_id, system_id)
            universe = fo.getUniverse()
            debug(
                "fleet %d with order type(%s) sent to safe leg dest %s and ultimate dest %s"
                % (fleet_id, self.ORDER_NAME, universe.getSystem(dest_id), universe.getSystem(system_id))
            )
            fo.issueFleetMoveOrder(fleet_id, dest_id)
            debug(f"Order issued: {self.ORDER_NAME} fleet: {self.fleet} target: {self.target}")
        ships_cur_health, ships_max_health = FleetUtilsAI.get_current_and_max_structure(fleet_id)
        self.executed = ships_cur_health == ships_max_health
        return True


def trooper_move_reqs_met(invasion_target: Target, order: OrderMove, verbose: bool) -> bool:
    """
    Indicates whether or not move requirements specific to invasion troopers are met for the provided mission and order.

    :param verbose: whether to print verbose decision details
    """
    # Don't advance outside of our fleet-supply zone unless the target either has no shields at all or there
    # is already a military fleet assigned to secure the target, and don't take final jump unless the planet is
    # (to the AI's knowledge) down to zero shields.  Additional checks will also be done by the later
    # generic movement code
    invasion_planet = invasion_target.get_object()
    invasion_system = invasion_target.get_system()
    supplied_systems = fo.getEmpire().fleetSupplyableSystemIDs
    # if about to leave supply lines
    if order.target.id not in supplied_systems or fo.getUniverse().jumpDistance(order.fleet.id, invasion_system.id) < 5:
        if invasion_planet.currentMeterValue(fo.meterType.maxShield):
            military_support_fleets = MilitaryAI.get_military_fleets_with_target_system(invasion_system.id)
            # if there is a threat in the enemy system, do give military ships at least 1 turn to clear it
            delay_to_move_troops = 1 if MilitaryAI.get_system_local_threat(order.target.id) else 0

            def eta(fleet_id):
                return FleetUtilsAI.calculate_estimated_time_of_arrival(fleet_id, invasion_system.id)

            eta_this_fleet = eta(order.fleet.id)
            if all(
                ((eta_this_fleet - delay_to_move_troops) <= eta(fid) and eta(fid)) for fid in military_support_fleets
            ):
                if verbose:
                    debug(
                        "trooper_move_reqs_met() holding Invasion fleet %d before leaving supply "
                        "because target (%s) has nonzero max shields and no assigned military fleet would arrive"
                        "at least %d turn earlier than the invasion fleet"
                        % (order.fleet.id, invasion_planet, delay_to_move_troops)
                    )
                return False

        if verbose:
            debug(
                "trooper_move_reqs_met() allowing Invasion fleet %d to leave supply "
                "because target (%s) has zero max shields or there is a military fleet assigned to secure "
                "the target system which will arrive at least 1 turn before the invasion fleet.",
                order.fleet.id,
                invasion_planet,
            )
    return True



from collections.abc import Sequence

from common.fo_typing import (
    BuildingId,
    BuildingName,
    EmpireId,
    FleetId,
    ObjectId,
    PartName,
    PlanetId,
    PlayerId,
    ShipId,
    SpeciesName,
    SystemId,
    Turn,
)
from stub_generator.stub_generator.collection_classes import make_type
from stub_generator.stub_generator.rtype.mapper import Mapper
from stub_generator.stub_generator.rtype.utils import get_name_for_mapping

_property_map = Mapper(
    "property",
    {
        ("shipIDs", "IntVec"): Sequence[ShipId],
        ("buildingIDs", "IntVec"): Sequence[BuildingId],
        ("planetIDs", "IntVec"): Sequence[PlanetId],
        ("fleetIDs", "IntVec"): Sequence[FleetId],
        ("systemIDs", "IntVec"): Sequence[SystemId],
        ("empireID", "int"): EmpireId,
        ("capitalID", "int"): PlanetId,
        ("locationID", "int"): PlanetId,
        ("speciesName", "str"): SpeciesName,
        ("designedOnTurn", "int"): Turn,
        ("buildingTypeName", "str"): BuildingName,
        ("parts", "StringVec"): Sequence[PartName],
        ("recipient", "int"): PlayerId,
        ("sender", "int"): PlayerId,
        ("id", ""): ObjectId,
        ("systemID", ""): SystemId,
        ("name", ""): str,
        ("owner", ""): EmpireId,
        ("empire1", ""): EmpireId,
        ("empire2", ""): EmpireId,
        ("status", ""): "diplomaticStatus",
        ("specials", ""): SpeciesName,
        ("producedByEmpireID", "int"): EmpireId,
        ("planetID", "int"): PlanetId,
    },
)


def update_property_rtype(attr_name: str, rtype: str):
    if rtype.startswith("<type"):
        return rtype[7:-2]

    rtype = rtype.split(".")[-1].strip("'>")

    key = (attr_name, rtype)
    if key in _property_map:
        property_class = _property_map[key]

        return get_name_for_mapping(property_class)
    else:
        return make_type(rtype)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="GRO_NANOTECH_MED",
    description="GRO_NANOTECH_MED_DESC",
    short_description="POLICY_UNLOCK_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=75 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_GROWTH_CATEGORY", "THEORY"],
    prerequisites=[
        "GRO_GENETIC_MED",
        "PRO_NANOTECH_PROD",
    ],
    unlock=Item(type=UnlockPolicy, name="PLC_POPULATION"),
    graphic="icons/tech/nanotech_medicine.png",
)



# comments hold info about internal engine rules defined in c++ side, but this was manually created in some effort to bring the two together
# and holds no power over what is on the C++ side
# for proper ordering ranks need to make sense together with GameRuleRanks.h
# categories are ranked by definitions in GameRuleCategories.h - add new category there in expected order, along with stringtable entry


def count(start=0, step=1):
    n = start
    while True:
        yield n
        n += step


#########################################
# Category "" aka "GENERAL"
#########################################
# -> only internal rules so far
#   internal RULE_RESEED_PRNG_SERVER
#   internal RULE_EXTRASOLAR_SHIP_DETECTION
#   internal RULE_BASIC_VIS_SYSTEM_INFO_SHOWN
#   internal RULE_NUM_COMBAT_ROUNDS
#   internal RULE_AGGRESSIVE_SHIPS_COMBAT_VISIBLE
#   internal RULE_STOCKPILE_IMPORT_LIMITED
#   internal RULE_PRODUCTION_QUEUE_FRONTLOAD_FACTOR
#   internal RULE_PRODUCTION_QUEUE_TOPPING_UP_FACTOR


#########################################
# Category "CONTENT"
#########################################
ranker = count(start=100, step=100)

GameRule(
    name="RULE_ALLOW_REPEATED_SPECIES",
    description="RULE_ALLOW_REPEATED_SPECIES_DESC",
    category="CONTENT",
    type=bool,
    default=False,
    rank=next(ranker),
)

GameRule(
    name="RULE_ENSURE_HABITABLE_PLANET_HW_VICINITY",
    description="RULE_ENSURE_HABITABLE_PLANET_HW_VICINITY_DESC",
    category="CONTENT",
    type=bool,
    default=False,
    rank=next(ranker),
)

GameRule(
    name="RULE_ENABLE_EXPERIMENTORS",
    description="RULE_ENABLE_EXPERIMENTORS_DESC",
    category="CONTENT",
    type=bool,
    default=True,
    rank=next(ranker),
)

GameRule(
    name="RULE_EXPERIMENTORS_SPAWN_BASE_TURN",
    description="RULE_EXPERIMENTORS_SPAWN_BASE_TURN_DESC",
    category="CONTENT",
    type=int,
    default=250,
    min=0,
    max=500,
    rank=next(ranker),
)

GameRule(
    name="RULE_NESTS_ALWAYS_SPAWN_WILD",
    description="RULE_NESTS_ALWAYS_SPAWN_WILD_DESC",
    category="CONTENT",
    type=bool,
    default=False,
    rank=next(ranker),
)

GameRule(
    name="RULE_WANDERING_MONSTERS_CAN_STOP_INSYSTEM",
    description="RULE_WANDERING_MONSTERS_CAN_STOP_INSYSTEM_DESC",
    category="CONTENT",
    type=bool,
    default=False,
    rank=next(ranker),
)

GameRule(
    name="RULE_WILD_NEST_MONSTER_SPAWN_FACTOR",
    description="RULE_WILD_NEST_MONSTER_SPAWN_FACTOR_DESC",
    category="CONTENT",
    type=float,
    default=1.0,
    min=0.0,
    max=10.0,
    rank=next(ranker),
)

GameRule(
    name="RULE_DOMESTIC_NEST_MONSTER_SPAWN_FACTOR",
    description="RULE_DOMESTIC_NEST_MONSTER_SPAWN_FACTOR_DESC",
    category="CONTENT",
    type=float,
    default=1.0,
    min=0.0,
    max=10.0,
    rank=next(ranker),
)

#########################################
# Category "BALANCE"
#########################################
ranker = count(start=100, step=100)

GameRule(
    name="RULE_SHIP_HULL_COST_FACTOR",
    description="RULE_SHIP_HULL_COST_FACTOR_DESC",
    category="BALANCE",
    type=float,
    default=1.0,
    min=0.1,
    max=10.0,
    rank=next(ranker),
)

GameRule(
    name="RULE_SHIP_PART_COST_FACTOR",
    description="RULE_SHIP_PART_COST_FACTOR_DESC",
    category="BALANCE",
    type=float,
    default=1.0,
    min=0.1,
    max=10.0,
    rank=next(ranker),
)

GameRule(
    name="RULE_BUILDING_COST_FACTOR",
    description="RULE_BUILDING_COST_FACTOR_DESC",
    category="BALANCE",
    type=float,
    default=1.0,
    min=0.1,
    max=10.0,
    rank=next(ranker),
)

GameRule(
    name="RULE_TECH_COST_FACTOR",
    description="RULE_TECH_COST_FACTOR_DESC",
    category="BALANCE",
    type=float,
    default=1.0,
    min=0.1,
    max=10.0,
    rank=next(ranker),
)

GameRule(
    name="RULE_SINGULARITY_COST_FACTOR",
    description="RULE_SINGULARITY_COST_FACTOR_DESC",
    category="BALANCE",
    type=float,
    default=1.0,
    min=0.1,
    max=10.0,
    rank=next(ranker),
)

GameRule(
    name="RULE_MINIMUM_MONSTER_DISTANCE_CAPITAL",
    description="RULE_MINIMUM_MONSTER_DISTANCE_CAPITAL_DESC",
    category="BALANCE",
    type=int,
    default=3,
    min=1,
    max=10,
    rank=next(ranker),
)

GameRule(
    name="RULE_SHIP_PART_BASED_UPKEEP",
    description="RULE_SHIP_PART_BASED_UPKEEP_DESC",
    category="BALANCE",
    type=bool,
    default=False,
    rank=next(ranker),
)

GameRule(
    name="RULE_FIRST_COMBAT_ROUND_IN_CLOSE_TARGETING_RANGE",
    description="RULE_FIRST_COMBAT_ROUND_IN_CLOSE_TARGETING_RANGE_DESC",
    category="BALANCE",
    type=int,
    default=3,
    min=1,
    max=20,
    rank=next(ranker),
)

# internal RULE_SHIP_WEAPON_DAMAGE_FACTOR rank = 1100,
# internal RULE_FIGHTER_DAMAGE_FACTOR rank = 1200,
# internal RULE_SHIP_STRUCTURE_FACTOR rank = 1300,
# internal RULE_SHIP_SPEED_FACTOR rank = 1400,


#########################################
# category "TEST"
#########################################
ranker = count(start=100, step=100)

GameRule(
    name="RULE_ENABLE_SUPER_TESTER",
    description="RULE_ENABLE_SUPER_TESTER_DESC",
    category="TEST",
    type=bool,
    default=False,
    rank=next(ranker),
)

# internal RULE_CHEAP_AND_FAST_BUILDING_PRODUCTION, rank = 110,
# internal RULE_CHEAP_AND_FAST_SHIP_PRODUCTION, rank = 111,
# internal RULE_CHEAP_AND_FAST_TECH_RESEARCH, rank = 112,
# internal RULE_CHEAP_POLICIES, rank = 113,
# internal RULE_ALL_OBJECTS_VISIBLE, rank = 130,
# internal RULE_ALL_SYSTEMS_VISIBLE, rank = 131,
# internal RULE_UNSEEN_STEALTHY_PLANETS_INVISIBLE, rank = 132,
# internal RULE_STARLANES_EVERYWHERE, rank = 135,

GameRule(
    name="RULE_TEST_STRING",
    description="RULE_TEST_STRING_DESC",
    category="TEST",
    type=str,
    default="PLAYER",
    allowed=["MODERATOR", "OBSERVER", "PLAYER", "AI_PLAYER"],
    rank=next(ranker),
)


#########################################
# category "BALANCE_STABILITY"
#########################################
ranker = count(start=100, step=100)

GameRule(
    name="RULE_BASELINE_PLANET_STABILITY",
    description="RULE_BASELINE_PLANET_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=0,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_PROTECTION_FOCUS_STABILITY",
    description="RULE_PROTECTION_FOCUS_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=15,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_IMPERIAL_PALACE_INFLUENCE",
    description="RULE_IMPERIAL_PALACE_INFLUENCE_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=3,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_GOOD_ENVIRONMENT_STABILITY",
    description="RULE_GOOD_ENVIRONMENT_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=2,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_ADEQUATE_ENVIRONMENT_STABILITY",
    description="RULE_ADEQUATE_ENVIRONMENT_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=1,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_POOR_ENVIRONMENT_STABILITY",
    description="RULE_POOR_ENVIRONMENT_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=0,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_HOSTILE_ENVIRONMENT_STABILITY",
    description="RULE_HOSTILE_ENVIRONMENT_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=-1,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_TINY_SIZE_STABILITY",
    description="RULE_TINY_SIZE_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=2,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_SMALL_SIZE_STABILITY",
    description="RULE_SMALL_SIZE_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=1,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_MEDIUM_SIZE_STABILITY",
    description="RULE_MEDIUM_SIZE_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=0,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_LARGE_SIZE_STABILITY",
    description="RULE_LARGE_SIZE_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=-1,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_HUGE_SIZE_STABILITY",
    description="RULE_HUGE_SIZE_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=-2,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_GAS_GIANT_SIZE_STABILITY",
    description="RULE_GAS_GIANT_SIZE_STABILITY_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=0,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_BASELINE_SPECIES_EMPIRE_OPINION",
    description="RULE_BASELINE_SPECIES_EMPIRE_OPINION_DESC",
    category="BALANCE_STABILITY",
    type=int,
    default=0,
    min=-20,
    max=20,
    rank=next(ranker),
)

GameRule(
    name="RULE_INVASION_OPINION_PENALTY_SCALING",
    description="RULE_INVASION_OPINION_PENALTY_SCALING_DESC",
    category="BALANCE_STABILITY",
    type=float,
    default=-2.0,
    min=-50.0,
    max=25.0,
    rank=next(ranker),
)

GameRule(
    name="RULE_SHIPS_LOST_DESTROYED_PENALTY_SCALING",
    description="RULE_SHIPS_LOST_DESTROYED_PENALTY_SCALING_DESC",
    category="BALANCE_STABILITY",
    type=float,
    default=-0.5,
    min=-50.0,
    max=25.0,
    rank=next(ranker),
)

GameRule(
    name="RULE_COLONIES_FOUNDED_BONUS_SCALING",
    description="RULE_COLONIES_FOUNDED_BONUS_SCALING_DESC",
    category="BALANCE_STABILITY",
    type=float,
    default=1.0,
    min=-25.0,
    max=50.0,
    rank=next(ranker),
)

# internal RULE_ANNEX_COST_OPINION_EXP_BASE, rank = 5500,
# internal RULE_ANNEX_COST_STABILITY_EXP_BASE, rank = 5510,
# internal RULE_ANNEX_COST_SCALING, rank = 5520,
# internal RULE_BUILDING_ANNEX_COST_SCALING, rank = 5530,
# internal RULE_ANNEX_COST_MINIMUM, rank = 5540,

#########################################
# category "PLANET_SIZE"
#########################################
ranker = count(start=100, step=100)

GameRule(
    name="RULE_HABITABLE_SIZE_TINY",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=1,
    min=0,
    max=999,
    rank=next(ranker),
)

GameRule(
    name="RULE_HABITABLE_SIZE_SMALL",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=2,
    min=0,
    max=999,
    rank=next(ranker),
)

GameRule(
    name="RULE_HABITABLE_SIZE_MEDIUM",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=3,
    min=0,
    max=999,
    rank=next(ranker),
)

GameRule(
    name="RULE_HABITABLE_SIZE_LARGE",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=4,
    min=0,
    max=999,
    rank=next(ranker),
)

GameRule(
    name="RULE_HABITABLE_SIZE_HUGE",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=5,
    min=0,
    max=999,
    rank=next(ranker),
)

GameRule(
    name="RULE_HABITABLE_SIZE_ASTEROIDS",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=3,
    min=0,
    max=999,
    rank=next(ranker),
)

GameRule(
    name="RULE_HABITABLE_SIZE_GASGIANT",
    description="RULE_HABITABLE_SIZE_DESC",
    category="PLANET_SIZE",
    type=int,
    default=6,
    min=0,
    max=999,
    rank=next(ranker),
)


#########################################
# Category "MULTIPLAYER"
#########################################
# -> only internal rules so far
# internal RULE_DIPLOMACY, rank = 90000000,
# internal RULE_THRESHOLD_HUMAN_PLAYER_WIN, rank = 90000010,
# internal RULE_ONLY_ALLIANCE_WIN, rank = 90000020,
# internal RULE_ALLOW_CONCEDE, rank = 90000030,
# internal RULE_CONCEDE_COLONIES_THRESHOLD, rank = 90000040,



import freeOrionAIInterface as fo
from logging import debug, info

from common.print_utils import Sequence, Table, Text
from EnumsAI import EmpireProductionTypes
from turn_state import get_all_empire_planets


def print_building_list():
    debug("Buildings present on all owned planets:")
    universe = fo.getUniverse()
    for pid in get_all_empire_planets():
        planet = universe.getPlanet(pid)
        if planet:
            debug("%30s: %s" % (planet.name, [universe.getBuilding(bldg).name for bldg in planet.buildingIDs]))
    debug("")


def print_production_queue(after_turn=False):
    """Print production queue content with relevant info in table format."""
    universe = fo.getUniverse()
    s = "after" if after_turn else "before"
    title = "Production Queue Turn %d %s ProductionAI calls" % (fo.currentTurn(), s)
    prod_queue_table = Table(
        Text("Object"),
        Text("Location"),
        Text("Quantity"),
        Text("Progress"),
        Text("Allocated PP"),
        Text("Turns left"),
        table_name=title,
    )
    for element in fo.getEmpire().productionQueue:
        if element.buildType == EmpireProductionTypes.BT_SHIP:
            item = fo.getShipDesign(element.designID)
        elif element.buildType == EmpireProductionTypes.BT_BUILDING:
            item = fo.getBuildingType(element.name)
        else:
            continue
        cost = item.productionCost(fo.empireID(), element.locationID)

        prod_queue_table.add_row(
            element.name,
            universe.getPlanet(element.locationID),
            "%dx %d" % (element.remaining, element.blocksize),
            f"{element.progress * cost:.1f} / {cost:.1f}",
            "%.1f" % element.allocation,
            "%d" % element.turnsLeft,
        )
    info(prod_queue_table)


def print_capital_info(homeworld):
    table = Table(
        Text("Id", description="Building id"),
        Text("Name"),
        Text("Type"),
        Sequence("Tags"),
        Sequence("Specials"),
        Text("Owner Id"),
        table_name="Buildings present at empire Capital in Turn %d" % fo.currentTurn(),
    )

    universe = fo.getUniverse()

    for building_id in homeworld.buildingIDs:
        building = universe.getBuilding(building_id)

        table.add_row(
            building_id,
            building.name,
            "_".join(building.buildingTypeName.split("_")[-2:]),
            sorted(building.tags),
            sorted(building.specials),
            building.owner,
        )

    info(table)



import freeOrionAIInterface as fo

from CombatRatingsAI._ship_combat_stats import ShipCombatStats, get_ship_combat_stats
from common.fo_typing import FleetId
from freeorion_tools import combine_ratings
from freeorion_tools.caching import cache_for_current_turn


class FleetCombatStats:
    """Stores combat related stats of the fleet."""

    def __init__(self, fleet_id: FleetId, max_stats=False):
        self._fleet_id = fleet_id
        self._max_stats = max_stats
        self._ship_stats = self._get_stats_from_fleet()

    def get_ship_combat_stats(self) -> list[ShipCombatStats]:
        """Returns list of ShipCombatStats of the fleet."""
        return list(self._ship_stats)

    def get_rating(self, enemy_stats: ShipCombatStats = None) -> float:
        """Calculates the rating of the fleet by combining all its ships ratings.

        :param enemy_stats: enemy to be rated against
        :return: Rating of the fleet
        """
        return combine_ratings(x.get_rating(enemy_stats) for x in self._ship_stats)

    def get_rating_vs_planets(self) -> float:
        return combine_ratings(x.get_rating_vs_planets() for x in self._ship_stats)

    def _get_stats_from_fleet(self):
        """Calculate fleet combat stats (i.e. the stats of all its ships)."""
        universe = fo.getUniverse()
        fleet = universe.getFleet(self._fleet_id)
        if not fleet:
            return []
        return [get_ship_combat_stats(ship_id=ship_id, max_stats=self._max_stats) for ship_id in fleet.shipIDs]


@cache_for_current_turn
def get_fleet_combat_stats(fleet_id, max_stats=False):
    return FleetCombatStats(fleet_id, max_stats=max_stats)


def get_ships_stats_for_fleet(fleet_id, max_stats=False) -> list[ShipCombatStats]:
    return get_fleet_combat_stats(fleet_id, max_stats=max_stats).get_ship_combat_stats()



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_NOVA_BOMB",
    description="SHP_NOVA_BOMB_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=960 * TECH_COST_MULTIPLIER,
    researchturns=12,
    tags=["PEDIA_SHIP_WEAPONS_CATEGORY"],
    prerequisites=["LRN_STELLAR_TOMOGRAPHY", "PRO_ZERO_GEN"],
    unlock=[Item(type=UnlockShipPart, name="SP_NOVA_BOMB"), Item(type=UnlockBuilding, name="BLD_NOVA_BOMB_ACTIVATOR")],
    graphic="icons/ship_parts/nova-bomb.png",
)
